import json
import logging
import re
import string
import uuid
import warnings
from asyncio import Lock
from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager
from copy import copy
from datetime import UTC, datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    Union,
    overload,
)

import sqlalchemy as sa
from slixmpp import JID, Iq, Message, Presence
from slixmpp.exceptions import IqError, IqTimeout, XMPPError
from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0060.stanza import Item
from slixmpp.plugins.xep_0082 import parse as str_to_datetime
from slixmpp.plugins.xep_0469.stanza import NS as PINNING_NS
from slixmpp.plugins.xep_0492.stanza import NS as NOTIFY_NS
from slixmpp.plugins.xep_0492.stanza import WhenLiteral
from slixmpp.xmlstream import ET
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session as OrmSession

from ..contact.contact import LegacyContact
from ..contact.roster import ContactIsUser
from ..core.mixins.avatar import AvatarMixin
from ..core.mixins.disco import ChatterDiscoMixin
from ..core.mixins.recipient import ReactionRecipientMixin, ThreadRecipientMixin
from ..db.models import Participant, Room
from ..util.archive_msg import HistoryMessage
from ..util.jid_escaping import unescape_node
from ..util.types import (
    HoleBound,
    LegacyGroupIdType,
    LegacyMessageType,
    LegacyParticipantType,
    LegacyThreadType,
    LegacyUserIdType,
    Mention,
    MucAffiliation,
    MucType,
)
from ..util.util import SubclassableOnce, deprecated, timeit
from .archive import MessageArchive
from .participant import LegacyParticipant, escape_nickname

if TYPE_CHECKING:
    from ..core.session import BaseSession

ADMIN_NS = "http://jabber.org/protocol/muc#admin"

SubjectSetterType = Union[str, None, "LegacyContact", "LegacyParticipant"]


class LegacyMUC(
    Generic[
        LegacyGroupIdType, LegacyMessageType, LegacyParticipantType, LegacyUserIdType
    ],
    AvatarMixin,
    ChatterDiscoMixin,
    ReactionRecipientMixin,
    ThreadRecipientMixin,
    SubclassableOnce,
):
    """
    A room, a.k.a. a Multi-User Chat.

    MUC instances are obtained by calling :py:meth:`slidge.group.bookmarks.LegacyBookmarks`
    on the user's :py:class:`slidge.core.session.BaseSession`.
    """

    max_history_fetch = 100

    is_group = True

    DISCO_TYPE = "text"
    DISCO_CATEGORY = "conference"

    STABLE_ARCHIVE = False
    """
    Because legacy events like reactions, editions, etc. don't all map to a stanza
    with a proper legacy ID, slidge usually cannot guarantee the stability of the archive
    across restarts.

    Set this to True if you know what you're doing, but realistically, this can't
    be set to True until archive is permanently stored on disk by slidge.

    This is just a flag on archive responses that most clients ignore anyway.
    """

    _ALL_INFO_FILLED_ON_STARTUP = False
    """
    Set this to true if the fill_participants() / fill_participants() design does not
    fit the legacy API, ie, no lazy loading of the participant list and history.
    """

    HAS_DESCRIPTION = True
    """
    Set this to false if the legacy network does not allow setting a description
    for the group. In this case the description field will not be present in the
    room configuration form.
    """

    HAS_SUBJECT = True
    """
    Set this to false if the legacy network does not allow setting a subject
    (sometimes also called topic) for the group. In this case, as a subject is
    recommended by :xep:`0045` ("SHALL"), the description (or the group name as
    ultimate fallback) will be used as the room subject.
    By setting this to false, an error will be returned when the :term:`User`
    tries to set the room subject.
    """

    archive: MessageArchive
    session: "BaseSession"

    stored: Room

    _participant_cls: type[LegacyParticipantType]

    def __init__(self, session: "BaseSession", stored: Room) -> None:
        self.session = session
        self.xmpp = session.xmpp
        self.stored = stored
        self._set_logger()
        super().__init__()

        self.archive = MessageArchive(stored, self.xmpp.store)

        if self._ALL_INFO_FILLED_ON_STARTUP:
            self.stored.participants_filled = True

    def pop_unread_xmpp_ids_up_to(self, horizon_xmpp_id: str) -> list[str]:
        """
        Return XMPP msg ids sent in this group up to a given XMPP msg id.

        Plugins have no reason to use this, but it is used by slidge core
        for legacy networks that need to mark *all* messages as read (most XMPP
        clients only send a read marker for the latest message).

        This has side effects: all messages up to the horizon XMPP id will be marked
        as read in the DB. If the horizon XMPP id is not found, all messages of this
        MUC will be marked as read.

        :param horizon_xmpp_id: The latest message
        :return: A list of XMPP ids if horizon_xmpp_id was not found
        """
        with self.xmpp.store.session() as orm:
            assert self.stored.id is not None
            ids = self.xmpp.store.mam.pop_unread_up_to(
                orm, self.stored.id, horizon_xmpp_id
            )
            orm.commit()
        return ids

    def participant_from_store(
        self, stored: Participant, contact: LegacyContact | None = None
    ) -> LegacyParticipantType:
        if contact is None and stored.contact is not None:
            contact = self.session.contacts.from_store(stored.contact)
        return self._participant_cls(self, stored=stored, contact=contact)

    @property
    def jid(self) -> JID:
        return self.stored.jid

    @jid.setter
    def jid(self, x: JID):
        # FIXME: without this, mypy yields
        #        "Cannot override writeable attribute with read-only property"
        #        But it does not happen for LegacyContact. WTF?
        raise RuntimeError

    @property
    def legacy_id(self):
        return self.xmpp.LEGACY_ROOM_ID_TYPE(self.stored.legacy_id)

    def orm(self, **kwargs) -> OrmSession:
        return self.xmpp.store.session(**kwargs)

    @property
    def type(self) -> MucType:
        return self.stored.muc_type

    @type.setter
    def type(self, type_: MucType) -> None:
        if self.type == type_:
            return
        self.update_stored_attribute(muc_type=type_)

    @property
    def n_participants(self):
        return self.stored.n_participants

    @n_participants.setter
    def n_participants(self, n_participants: int | None) -> None:
        if self.stored.n_participants == n_participants:
            return
        self.update_stored_attribute(n_participants=n_participants)

    @property
    def user_jid(self):
        return self.session.user_jid

    def _set_logger(self) -> None:
        self.log = logging.getLogger(f"{self.user_jid}:muc:{self}")

    def __repr__(self) -> str:
        return f"<MUC #{self.stored.id} '{self.name}' ({self.stored.legacy_id} - {self.jid.user})'>"

    @property
    def subject_date(self) -> datetime | None:
        if self.stored.subject_date is None:
            return None
        return self.stored.subject_date.replace(tzinfo=UTC)

    @subject_date.setter
    def subject_date(self, when: datetime | None) -> None:
        if self.subject_date == when:
            return
        self.update_stored_attribute(subject_date=when)

    def __send_configuration_change(self, codes) -> None:
        part = self.get_system_participant()
        part.send_configuration_change(codes)

    @property
    def user_nick(self):
        return (
            self.stored.user_nick
            or self.session.bookmarks.user_nick
            or self.user_jid.node
        )

    @user_nick.setter
    def user_nick(self, nick: str) -> None:
        if nick == self.user_nick:
            return
        self.update_stored_attribute(user_nick=nick)

    def add_user_resource(self, resource: str) -> None:
        stored_set = self.get_user_resources()
        if resource in stored_set:
            return
        stored_set.add(resource)
        self.update_stored_attribute(
            user_resources=(json.dumps(list(stored_set)) if stored_set else None)
        )

    def get_user_resources(self) -> set[str]:
        stored_str = self.stored.user_resources
        if stored_str is None:
            return set()
        return set(json.loads(stored_str))

    def remove_user_resource(self, resource: str) -> None:
        stored_set = self.get_user_resources()
        if resource not in stored_set:
            return
        stored_set.remove(resource)
        self.update_stored_attribute(
            user_resources=(json.dumps(list(stored_set)) if stored_set else None)
        )

    @asynccontextmanager
    async def lock(self, id_: str) -> AsyncIterator[None]:
        async with self.session.lock((self.legacy_id, id_)):
            yield

    def get_lock(self, id_: str) -> Lock | None:
        return self.session.get_lock((self.legacy_id, id_))

    async def __fill_participants(self) -> None:
        if self._ALL_INFO_FILLED_ON_STARTUP or self.participants_filled:
            return

        async with self.lock("fill participants"):
            with self.xmpp.store.session(expire_on_commit=False) as orm:
                orm.add(self.stored)
                with orm.no_autoflush:
                    orm.refresh(self.stored, ["participants_filled"])
            if self.participants_filled:
                return
            parts: list[Participant] = []
            resources = set[str]()
            # During fill_participants(), self.get_participant*() methods may
            # return a participant with a conflicting nick/resource.
            async for participant in self.fill_participants():
                if participant.stored.resource in resources:
                    self.log.debug(
                        "Participant '%s' was yielded more than once by fill_participants(), ignoring",
                        participant.stored.resource,
                    )
                    continue
                parts.append(participant.stored)
                resources.add(participant.stored.resource)
        with self.xmpp.store.session(expire_on_commit=False) as orm:
            orm.add(self.stored)
            # because self.fill_participants() is async, self.stored may be stale at
            # this point, and the only thing we want to update is the participant list
            # and the participant_filled attribute.
            with orm.no_autoflush:
                orm.refresh(self.stored)
                for part in parts:
                    orm.merge(part)
            self.stored.participants_filled = True
            orm.commit()

    async def get_participants(
        self, affiliation: MucAffiliation | None = None
    ) -> AsyncIterator[LegacyParticipantType]:
        await self.__fill_participants()
        with self.xmpp.store.session(expire_on_commit=False, autoflush=False) as orm:
            self.stored = orm.merge(self.stored)
            for db_participant in self.stored.participants:
                if (
                    affiliation is not None
                    and db_participant.affiliation != affiliation
                ):
                    continue
                yield self.participant_from_store(db_participant)

    async def __fill_history(self) -> None:
        async with self.lock("fill history"):
            with self.xmpp.store.session(expire_on_commit=False) as orm:
                orm.add(self.stored)
                with orm.no_autoflush:
                    orm.refresh(self.stored, ["history_filled"])
            if self.stored.history_filled:
                self.log.debug("History has already been fetched.")
                return
            log.debug("Fetching history for %s", self)
            try:
                before, after = self.archive.get_hole_bounds()
                if before is not None:
                    before = before._replace(
                        id=self.xmpp.LEGACY_MSG_ID_TYPE(before.id)  # type:ignore
                    )
                if after is not None:
                    after = after._replace(
                        id=self.xmpp.LEGACY_MSG_ID_TYPE(after.id)  # type:ignore
                    )
                await self.backfill(before, after)
            except NotImplementedError:
                return
            except Exception as e:
                self.log.exception("Could not backfill", exc_info=e)

            self.stored.history_filled = True
            self.commit(merge=True)

    def _get_disco_name(self) -> str | None:
        return self.name

    @property
    def name(self) -> str | None:
        return self.stored.name

    @name.setter
    def name(self, n: str | None) -> None:
        if self.name == n:
            return
        self.update_stored_attribute(name=n)
        self._set_logger()
        self.__send_configuration_change((104,))

    @property
    def description(self):
        return self.stored.description or ""

    @description.setter
    def description(self, d: str) -> None:
        if self.description == d:
            return
        self.update_stored_attribute(description=d)
        self.__send_configuration_change((104,))

    def on_presence_unavailable(self, p: Presence) -> None:
        pto = p.get_to()
        if pto.bare != self.jid.bare:
            return

        pfrom = p.get_from()
        if pfrom.bare != self.user_jid.bare:
            return
        if (resource := pfrom.resource) in self.get_user_resources():
            if pto.resource != self.user_nick:
                self.log.debug(
                    "Received 'leave group' request but with wrong nickname. %s", p
                )
            self.remove_user_resource(resource)
        else:
            self.log.debug(
                "Received 'leave group' request but resource was not listed. %s", p
            )

    async def update_info(self):
        """
        Fetch information about this group from the legacy network

        This is awaited on MUC instantiation, and should be overridden to
        update the attributes of the group chat, like title, subject, number
        of participants etc.

        To take advantage of the slidge avatar cache, you can check the .avatar
        property to retrieve the "legacy file ID" of the cached avatar. If there
        is no change, you should not call
        :py:meth:`slidge.core.mixins.avatar.AvatarMixin.set_avatar()` or
        attempt to modify
        the :attr:.avatar property.
        """
        raise NotImplementedError

    async def backfill(
        self,
        after: HoleBound | None = None,
        before: HoleBound | None = None,
    ):
        """
        Override this if the legacy network provide server-side group archives.

        In it, send history messages using ``self.get_participant(xxx).send_xxxx``,
        with the ``archive_only=True`` kwarg. This is only called once per slidge
        run for a given group.

        :param after: Fetch messages after this one.
            If ``None``, slidge's local archive was empty before start-up,
            ie, no history was ever fetched for this room since the user registered.
            It's up to gateway implementations to decide how far to fetch messages before
            the user registered.
            If not ``None``, slidge has some messages in this archive, and
            the gateway shall try to fetch history up to (and excluding) this message
            to avoid "holes" in the history of this group.
        :param before: Fetch messages before this one.
            If ``None``, the gateway shall fetch all messages up to the most recent one.
            If not ``None``, slidge has already archived some live messages
            it received during its lifetime, and there is no need to query the legacy
            network for any message after (and including) this one.
        """
        raise NotImplementedError

    async def fill_participants(self) -> AsyncIterator[LegacyParticipantType]:
        """
        This method should yield the list of all members of this group.

        Typically, use ``participant = self.get_participant()``, self.get_participant_by_contact(),
        of self.get_user_participant(), and update their affiliation, hats, etc.
        before yielding them.
        """
        return
        yield

    @property
    def subject(self) -> str:
        return self.stored.subject or ""

    @subject.setter
    def subject(self, s: str) -> None:
        if s == self.subject:
            return

        self.update_stored_attribute(subject=s)
        self.__get_subject_setter_participant().set_room_subject(
            s, None, self.subject_date, False
        )

    @property
    def is_anonymous(self):
        return self.type == MucType.CHANNEL

    @property
    def subject_setter(self) -> str | None:
        return self.stored.subject_setter

    @subject_setter.setter
    def subject_setter(self, subject_setter: SubjectSetterType) -> None:
        if isinstance(subject_setter, LegacyContact):
            subject_setter = subject_setter.name
        elif isinstance(subject_setter, LegacyParticipant):
            subject_setter = subject_setter.nickname

        if subject_setter == self.subject_setter:
            return
        assert isinstance(subject_setter, str | None)
        self.update_stored_attribute(subject_setter=subject_setter)

    def __get_subject_setter_participant(self) -> LegacyParticipant:
        if self.subject_setter is None:
            return self.get_system_participant()
        return self._participant_cls(
            self,
            Participant(nickname=self.subject_setter, occupant_id="subject-setter"),
        )

    def features(self):
        features = [
            "http://jabber.org/protocol/muc",
            "http://jabber.org/protocol/muc#stable_id",
            "http://jabber.org/protocol/muc#self-ping-optimization",
            "urn:xmpp:mam:2",
            "urn:xmpp:mam:2#extended",
            "urn:xmpp:sid:0",
            "muc_persistent",
            "vcard-temp",
            "urn:xmpp:ping",
            "urn:xmpp:occupant-id:0",
            "jabber:iq:register",
            self.xmpp.plugin["xep_0425"].stanza.NS,
        ]
        if self.type == MucType.GROUP:
            features.extend(["muc_membersonly", "muc_nonanonymous", "muc_hidden"])
        elif self.type == MucType.CHANNEL:
            features.extend(["muc_open", "muc_semianonymous", "muc_public"])
        elif self.type == MucType.CHANNEL_NON_ANONYMOUS:
            features.extend(["muc_open", "muc_nonanonymous", "muc_public"])
        return features

    async def extended_features(self):
        is_group = self.type == MucType.GROUP

        form = self.xmpp.plugin["xep_0004"].make_form(ftype="result")

        form.add_field(
            "FORM_TYPE", "hidden", value="http://jabber.org/protocol/muc#roominfo"
        )
        form.add_field("muc#roomconfig_persistentroom", "boolean", value=True)
        form.add_field("muc#roomconfig_changesubject", "boolean", value=False)
        form.add_field("muc#maxhistoryfetch", value=str(self.max_history_fetch))
        form.add_field("muc#roominfo_subjectmod", "boolean", value=False)

        if self.stored.id is not None and (
            self._ALL_INFO_FILLED_ON_STARTUP or self.stored.participants_filled
        ):
            with self.xmpp.store.session() as orm:
                n = orm.scalar(
                    sa.select(sa.func.count(Participant.id)).filter_by(
                        room_id=self.stored.id
                    )
                )
        else:
            n = self.n_participants
        if n is not None:
            form.add_field("muc#roominfo_occupants", value=str(n))

        if d := self.description:
            form.add_field("muc#roominfo_description", value=d)

        if s := self.subject:
            form.add_field("muc#roominfo_subject", value=s)

        if name := self.name:
            form.add_field("muc#roomconfig_roomname", value=name)

        if self._set_avatar_task is not None:
            await self._set_avatar_task
        avatar = self.get_avatar()
        if avatar and (h := avatar.id):
            form.add_field(
                "{http://modules.prosody.im/mod_vcard_muc}avatar#sha1", value=h
            )
            form.add_field("muc#roominfo_avatarhash", "text-multi", value=[h])

        form.add_field("muc#roomconfig_membersonly", "boolean", value=is_group)
        form.add_field(
            "muc#roomconfig_whois",
            "list-single",
            value="moderators" if self.is_anonymous else "anyone",
        )
        form.add_field("muc#roomconfig_publicroom", "boolean", value=not is_group)
        form.add_field("muc#roomconfig_allowpm", "boolean", value=False)

        r = [form]

        if reaction_form := await self.restricted_emoji_extended_feature():
            r.append(reaction_form)

        return r

    def shutdown(self) -> None:
        _, user_jid = escape_nickname(self.jid, self.user_nick)
        for user_full_jid in self.user_full_jids():
            presence = self.xmpp.make_presence(
                pfrom=user_jid, pto=user_full_jid, ptype="unavailable"
            )
            presence["muc"]["affiliation"] = "none"
            presence["muc"]["role"] = "none"
            presence["muc"]["status_codes"] = {110, 332}
            presence.send()

    def user_full_jids(self):
        for r in self.get_user_resources():
            j = JID(self.user_jid)
            j.resource = r
            yield j

    @property
    def user_muc_jid(self):
        _, user_muc_jid = escape_nickname(self.jid, self.user_nick)
        return user_muc_jid

    async def echo(
        self, msg: Message, legacy_msg_id: LegacyMessageType | None = None
    ) -> str:
        msg.set_from(self.user_muc_jid)
        if legacy_msg_id:
            msg["stanza_id"]["id"] = self.session.legacy_to_xmpp_msg_id(legacy_msg_id)
        else:
            msg["stanza_id"]["id"] = str(uuid.uuid4())
        msg["stanza_id"]["by"] = self.jid

        user_part = await self.get_user_participant()
        msg["occupant-id"]["id"] = user_part.stored.occupant_id

        self.archive.add(msg, user_part)

        for user_full_jid in self.user_full_jids():
            self.log.debug("Echoing to %s", user_full_jid)
            msg = copy(msg)
            msg.set_to(user_full_jid)

            msg.send()

        return msg["stanza_id"]["id"]

    def _post_avatar_update(self, cached_avatar) -> None:
        self.__send_configuration_change((104,))
        self._send_room_presence()

    def _send_room_presence(self, user_full_jid: JID | None = None) -> None:
        if user_full_jid is None:
            tos = self.user_full_jids()
        else:
            tos = [user_full_jid]
        for to in tos:
            p = self.xmpp.make_presence(pfrom=self.jid, pto=to)
            if (avatar := self.get_avatar()) and (h := avatar.id):
                p["vcard_temp_update"]["photo"] = h
            else:
                p["vcard_temp_update"]["photo"] = ""
            p.send()

    @timeit
    async def join(self, join_presence: Presence):
        user_full_jid = join_presence.get_from()
        requested_nickname = join_presence.get_to().resource
        client_resource = user_full_jid.resource

        if client_resource in self.get_user_resources():
            self.log.debug("Received join from a resource that is already joined.")

        if not requested_nickname or not client_resource:
            raise XMPPError("jid-malformed", by=self.jid)

        self.add_user_resource(client_resource)

        self.log.debug(
            "Resource %s of %s wants to join room %s with nickname %s",
            client_resource,
            self.user_jid,
            self.legacy_id,
            requested_nickname,
        )

        user_nick = self.user_nick
        user_participant = None
        async for participant in self.get_participants():
            if participant.is_user:
                user_participant = participant
                continue
            participant.send_initial_presence(full_jid=user_full_jid)

        if user_participant is None:
            user_participant = await self.get_user_participant()
            with self.xmpp.store.session() as orm:
                orm.add(self.stored)
                with orm.no_autoflush:
                    orm.refresh(self.stored, ["participants"])
        if not user_participant.is_user:
            self.log.warning("is_user flag not set on user_participant")
            user_participant.is_user = True
        user_participant.send_initial_presence(
            user_full_jid,
            presence_id=join_presence["id"],
            nick_change=user_nick != requested_nickname,
        )

        history_params = join_presence["muc_join"]["history"]
        maxchars = int_or_none(history_params["maxchars"])
        maxstanzas = int_or_none(history_params["maxstanzas"])
        seconds = int_or_none(history_params["seconds"])
        try:
            since = self.xmpp.plugin["xep_0082"].parse(history_params["since"])
        except ValueError:
            since = None
        if seconds is not None:
            since = datetime.now() - timedelta(seconds=seconds)
        if equals_zero(maxchars) or equals_zero(maxstanzas):
            log.debug("Joining client does not want any old-school MUC history-on-join")
        else:
            self.log.debug("Old school history fill")
            await self.__fill_history()
            await self.__old_school_history(
                user_full_jid,
                maxchars=maxchars,
                maxstanzas=maxstanzas,
                since=since,
            )
        if self.HAS_SUBJECT:
            subject = self.subject or ""
        else:
            subject = self.description or self.name or ""
        self.__get_subject_setter_participant().set_room_subject(
            subject,
            user_full_jid,
            self.subject_date,
        )
        if t := self._set_avatar_task:
            await t
        self._send_room_presence(user_full_jid)

    async def get_user_participant(self, **kwargs) -> "LegacyParticipantType":
        """
        Get the participant representing the gateway user

        :param kwargs: additional parameters for the :class:`.Participant`
            construction (optional)
        :return:
        """
        p = await self.get_participant(self.user_nick, is_user=True, **kwargs)
        self.__store_participant(p)
        return p

    def __store_participant(self, p: "LegacyParticipantType") -> None:
        if self.get_lock("fill participants"):
            return
        try:
            p.commit(merge=True)
        except IntegrityError as e:
            if self._ALL_INFO_FILLED_ON_STARTUP:
                log.debug("ℂould not store participant: %r", e)
                with self.orm(expire_on_commit=False) as orm:
                    self.stored = self.xmpp.store.rooms.get(
                        orm, self.user_pk, legacy_id=str(self.legacy_id)
                    )
                    p.stored.room = self.stored
                    orm.add(p.stored)
                    orm.commit()
            else:
                log.debug("ℂould not store participant: %r", e)

    @overload
    async def get_participant(self, nickname: str) -> "LegacyParticipantType": ...

    @overload
    async def get_participant(self, *, occupant_id: str) -> "LegacyParticipantType": ...

    @overload
    async def get_participant(
        self, *, occupant_id: str, create: Literal[False]
    ) -> "LegacyParticipantType | None": ...

    @overload
    async def get_participant(
        self, *, occupant_id: str, create: Literal[True]
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant(
        self, nickname: str, *, occupant_id: str
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant(
        self, nickname: str, *, create: Literal[False]
    ) -> "LegacyParticipantType | None": ...

    @overload
    async def get_participant(
        self, nickname: str, *, create: Literal[True]
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant(
        self,
        nickname: str,
        *,
        create: Literal[True],
        is_user: bool,
        fill_first: bool,
        store: bool,
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant(
        self,
        nickname: str,
        *,
        create: Literal[False],
        is_user: bool,
        fill_first: bool,
        store: bool,
    ) -> "LegacyParticipantType | None": ...

    @overload
    async def get_participant(
        self,
        nickname: str,
        *,
        create: bool,
        fill_first: bool,
    ) -> "LegacyParticipantType | None": ...

    async def get_participant(
        self,
        nickname: str | None = None,
        *,
        create: bool = True,
        is_user: bool = False,
        fill_first: bool = False,
        store: bool = True,
        occupant_id: str | None = None,
    ) -> "LegacyParticipantType | None":
        """
        Get a participant by their nickname.

        In non-anonymous groups, you probably want to use
        :meth:`.LegacyMUC.get_participant_by_contact` instead.

        :param nickname: Nickname of the participant (used as resource part in the MUC)
        :param create: By default, a participant is created if necessary. Set this to
            False to return None if participant was not created before.
        :param is_user: Whether this participant is the slidge user.
        :param fill_first: Ensure :meth:`.LegacyMUC.fill_participants()` has been called
            first (internal use by slidge, plugins should not need that)
        :param store: persistently store the user in the list of MUC participants
        :param occupant_id: optionally, specify the unique ID for this participant, cf
            xep:`0421`
        :return: A participant of this room.
        """
        if not any((nickname, occupant_id)):
            raise TypeError("You must specify either a nickname or an occupant ID")
        if fill_first:
            await self.__fill_participants()
        if not self._ALL_INFO_FILLED_ON_STARTUP or self.stored.id is not None:
            with self.xmpp.store.session(expire_on_commit=False) as orm:
                if occupant_id is not None:
                    stored = (
                        orm.query(Participant)
                        .filter(
                            Participant.room == self.stored,
                            Participant.occupant_id == occupant_id,
                        )
                        .one_or_none()
                    )
                elif nickname is not None:
                    stored = (
                        orm.query(Participant)
                        .filter(
                            Participant.room == self.stored,
                            (Participant.nickname == nickname)
                            | (Participant.resource == nickname),
                        )
                        .one_or_none()
                    )
                else:
                    raise RuntimeError("NEVER")
                if stored is not None:
                    if occupant_id and occupant_id != stored.occupant_id:
                        warnings.warn(
                            f"Occupant ID mismatch in get_participant(): {occupant_id} vs {stored.occupant_id}",
                        )
                    part = self.participant_from_store(stored)
                    if occupant_id and nickname and nickname != stored.nickname:
                        stored.nickname = nickname
                        orm.add(stored)
                        orm.commit()
                    return part

        if not create:
            return None

        if occupant_id is None:
            occupant_id = "slidge-user" if is_user else str(uuid.uuid4())

        if nickname is None:
            nickname = occupant_id

        if not self.xmpp.store.rooms.nick_available(orm, self.stored.id, nickname):
            nickname = f"{nickname} ({occupant_id})"
            if is_user:
                self.user_nick = nickname

        p = self._participant_cls(
            self,
            Participant(
                room=self.stored,
                nickname=nickname or occupant_id,
                is_user=is_user,
                occupant_id=occupant_id,
            ),
        )
        if store:
            self.__store_participant(p)
        if (
            not self.get_lock("fill participants")
            and not self.get_lock("fill history")
            and self.stored.participants_filled
            and not p.is_user
            and not p.is_system
        ):
            p.send_affiliation_change()
        return p

    def get_system_participant(self) -> "LegacyParticipantType":
        """
        Get a pseudo-participant, representing the room itself

        Can be useful for events that cannot be mapped to a participant,
        e.g. anonymous moderation events, or announces from the legacy
        service
        :return:
        """
        return self._participant_cls(
            self, Participant(occupant_id="room"), is_system=True
        )

    @overload
    async def get_participant_by_contact(
        self, c: "LegacyContact[Any]"
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant_by_contact(
        self, c: "LegacyContact[Any]", *, occupant_id: str | None = None
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant_by_contact(
        self,
        c: "LegacyContact[Any]",
        *,
        create: Literal[False],
        occupant_id: str | None,
    ) -> "LegacyParticipantType | None": ...

    @overload
    async def get_participant_by_contact(
        self,
        c: "LegacyContact[Any]",
        *,
        create: Literal[True],
        occupant_id: str | None,
    ) -> "LegacyParticipantType": ...

    async def get_participant_by_contact(
        self, c: "LegacyContact", *, create: bool = True, occupant_id: str | None = None
    ) -> "LegacyParticipantType | None":
        """
        Get a non-anonymous participant.

        This is what should be used in non-anonymous groups ideally, to ensure
        that the Contact jid is associated to this participant

        :param c: The :class:`.LegacyContact` instance corresponding to this contact
        :param create: Creates the participant if it does not exist.
        :param occupant_id: Optionally, specify a unique occupant ID (:xep:`0421`) for
            this participant.
        :return:
        """
        await self.session.contacts.ready

        if not self._ALL_INFO_FILLED_ON_STARTUP or self.stored.id is not None:
            with self.xmpp.store.session() as orm:
                self.stored = orm.merge(self.stored)
                stored = (
                    orm.query(Participant)
                    .filter_by(contact=c.stored, room=self.stored)
                    .one_or_none()
                )
                if stored is None:
                    if not create:
                        return None
                else:
                    if occupant_id and stored.occupant_id != occupant_id:
                        warnings.warn(
                            f"Occupant ID mismatch: {occupant_id} vs {stored.occupant_id}",
                        )
                    return self.participant_from_store(stored=stored, contact=c)

        nickname = c.name or unescape_node(c.jid.node)

        if self.stored.id is None:
            nick_available = True
        else:
            with self.xmpp.store.session() as orm:
                nick_available = self.xmpp.store.rooms.nick_available(
                    orm, self.stored.id, nickname
                )

        if not nick_available:
            self.log.debug("Nickname conflict")
            nickname = f"{nickname} ({c.jid.node})"
        p = self._participant_cls(
            self,
            Participant(
                nickname=nickname,
                room=self.stored,
                occupant_id=occupant_id or str(c.jid),
            ),
            contact=c,
        )

        self.__store_participant(p)
        # FIXME: this is not great but given the current design,
        #        during participants fill and history backfill we do not
        #        want to send presence, because we might :update affiliation
        #        and role afterwards.
        # We need a refactor of the MUC class… later™
        if (
            self.stored.participants_filled
            and not self.get_lock("fill participants")
            and not self.get_lock("fill history")
        ):
            p.send_last_presence(force=True, no_cache_online=True)
        return p

    @overload
    async def get_participant_by_legacy_id(
        self, legacy_id: LegacyUserIdType
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant_by_legacy_id(
        self,
        legacy_id: LegacyUserIdType,
        *,
        occupant_id: str | None,
        create: Literal[True],
    ) -> "LegacyParticipantType": ...

    @overload
    async def get_participant_by_legacy_id(
        self,
        legacy_id: LegacyUserIdType,
        *,
        occupant_id: str | None,
        create: Literal[False],
    ) -> "LegacyParticipantType | None": ...

    async def get_participant_by_legacy_id(
        self,
        legacy_id: LegacyUserIdType,
        *,
        occupant_id: str | None = None,
        create: bool = True,
    ) -> "LegacyParticipantType":
        try:
            c = await self.session.contacts.by_legacy_id(legacy_id)
        except ContactIsUser:
            return await self.get_user_participant(occupant_id=occupant_id)
        return await self.get_participant_by_contact(  # type:ignore[call-overload]
            c, create=create, occupant_id=occupant_id
        )

    def remove_participant(
        self,
        p: "LegacyParticipantType",
        kick: bool = False,
        ban: bool = False,
        reason: str | None = None,
    ):
        """
        Call this when a participant leaves the room

        :param p: The participant
        :param kick: Whether the participant left because they were kicked
        :param ban: Whether the participant left because they were banned
        :param reason: Optionally, a reason why the participant was removed.
        """
        if kick and ban:
            raise TypeError("Either kick or ban")
        with self.xmpp.store.session() as orm:
            orm.delete(p.stored)
            orm.commit()
        if kick:
            codes = {307}
        elif ban:
            codes = {301}
        else:
            codes = None
        presence = p._make_presence(ptype="unavailable", status_codes=codes)
        p.stored.affiliation = "outcast" if ban else "none"
        p.stored.role = "none"
        if reason:
            presence["muc"].set_item_attr("reason", reason)
        p._send(presence)

    def rename_participant(self, old_nickname: str, new_nickname: str) -> None:
        with self.xmpp.store.session() as orm:
            stored = (
                orm.query(Participant)
                .filter_by(room=self.stored, nickname=old_nickname)
                .one_or_none()
            )
            if stored is None:
                self.log.debug("Tried to rename a participant that we didn't know")
                return
            p = self.participant_from_store(stored)
            if p.nickname == old_nickname:
                p.nickname = new_nickname

    async def __old_school_history(
        self,
        full_jid: JID,
        maxchars: int | None = None,
        maxstanzas: int | None = None,
        seconds: int | None = None,
        since: datetime | None = None,
    ) -> None:
        """
        Old-style history join (internal slidge use)

        :param full_jid:
        :param maxchars:
        :param maxstanzas:
        :param seconds:
        :param since:
        :return:
        """
        if since is None:
            if seconds is None:
                start_date = datetime.now(tz=UTC) - timedelta(days=1)
            else:
                start_date = datetime.now(tz=UTC) - timedelta(seconds=seconds)
        else:
            start_date = since or datetime.now(tz=UTC) - timedelta(days=1)

        for h_msg in self.archive.get_all(
            start_date=start_date, end_date=None, last_page_n=maxstanzas
        ):
            msg = h_msg.stanza_component_ns
            msg["delay"]["stamp"] = h_msg.when
            msg.set_to(full_jid)
            self.xmpp.send(msg, False)

    async def send_mam(self, iq: Iq) -> None:
        await self.__fill_history()

        form_values = iq["mam"]["form"].get_values()

        start_date = str_to_datetime_or_none(form_values.get("start"))
        end_date = str_to_datetime_or_none(form_values.get("end"))

        after_id = form_values.get("after-id")
        before_id = form_values.get("before-id")

        sender = form_values.get("with")

        ids = form_values.get("ids") or ()

        if max_str := iq["mam"]["rsm"]["max"]:
            try:
                max_results = int(max_str)
            except ValueError:
                max_results = None
        else:
            max_results = None

        after_id_rsm = iq["mam"]["rsm"]["after"]
        after_id = after_id_rsm or after_id

        before_rsm = iq["mam"]["rsm"]["before"]
        if before_rsm is not None and max_results is not None:
            last_page_n = max_results
            # - before_rsm is True means the empty element <before />, which means
            #   "last page in chronological order", cf https://xmpp.org/extensions/xep-0059.html#backwards
            # - before_rsm == "an ID" means <before>an ID</before>
            if before_rsm is not True:
                before_id = before_rsm
        else:
            last_page_n = None

        first = None
        last = None
        count = 0

        it = self.archive.get_all(
            start_date,
            end_date,
            before_id,
            after_id,
            ids,
            last_page_n,
            sender,
            bool(iq["mam"]["flip_page"]),
        )

        for history_msg in it:
            last = xmpp_id = history_msg.id
            if first is None:
                first = xmpp_id

            wrapper_msg = self.xmpp.make_message(mfrom=self.jid, mto=iq.get_from())
            wrapper_msg["mam_result"]["queryid"] = iq["mam"]["queryid"]
            wrapper_msg["mam_result"]["id"] = xmpp_id
            wrapper_msg["mam_result"].append(history_msg.forwarded())

            wrapper_msg.send()
            count += 1

            if max_results and count == max_results:
                break

        if max_results:
            try:
                next(it)
            except StopIteration:
                complete = True
            else:
                complete = False
        else:
            complete = True

        reply = iq.reply()
        if not self.STABLE_ARCHIVE:
            reply["mam_fin"]["stable"] = "false"
        if complete:
            reply["mam_fin"]["complete"] = "true"
        reply["mam_fin"]["rsm"]["first"] = first
        reply["mam_fin"]["rsm"]["last"] = last
        reply["mam_fin"]["rsm"]["count"] = str(count)
        reply.send()

    async def send_mam_metadata(self, iq: Iq) -> None:
        await self.__fill_history()
        await self.archive.send_metadata(iq)

    async def kick_resource(self, r: str) -> None:
        """
        Kick a XMPP client of the user. (slidge internal use)

        :param r: The resource to kick
        """
        pto = JID(self.user_jid)
        pto.resource = r
        p = self.xmpp.make_presence(
            pfrom=(await self.get_user_participant()).jid, pto=pto
        )
        p["type"] = "unavailable"
        p["muc"]["affiliation"] = "none"
        p["muc"]["role"] = "none"
        p["muc"]["status_codes"] = {110, 333}
        p.send()

    async def __get_bookmark(self) -> Item | None:
        item = Item()
        item["id"] = self.jid

        iq = Iq(stype="get", sfrom=self.user_jid, sto=self.user_jid)
        iq["pubsub"]["items"]["node"] = self.xmpp["xep_0402"].stanza.NS
        iq["pubsub"]["items"].append(item)

        try:
            ans = await self.xmpp["xep_0356"].send_privileged_iq(iq)
            if len(ans["pubsub"]["items"]) != 1:
                return None
            # this below creates the item if it wasn't here already
            # (slixmpp annoying magic)
            item = ans["pubsub"]["items"]["item"]
            item["id"] = self.jid
            return item
        except IqTimeout:
            warnings.warn(f"Cannot fetch bookmark for {self.user_jid}: timeout")
            return None
        except IqError as exc:
            warnings.warn(f"Cannot fetch bookmark for {self.user_jid}: {exc}")
            return None
        except PermissionError:
            warnings.warn(
                "IQ privileges (XEP0356) are not set, we cannot fetch the user bookmarks"
            )
            return None

    async def add_to_bookmarks(
        self,
        auto_join: bool = True,
        preserve: bool = True,
        pin: bool | None = None,
        notify: WhenLiteral | None = None,
    ) -> None:
        """
        Add the MUC to the user's XMPP bookmarks (:xep:`0402')

        This requires that slidge has the IQ privileged set correctly
        on the XMPP server

        :param auto_join: whether XMPP clients should automatically join
            this MUC on startup. In theory, XMPP clients will receive
            a "push" notification when this is called, and they will
            join if they are online.
        :param preserve: preserve auto-join and bookmarks extensions
            set by the user outside slidge
        :param pin: Pin the group chat bookmark :xep:`0469`. Requires privileged entity.
            If set to ``None`` (default), the bookmark pinning status will be untouched.
        :param notify: Chat notification setting: :xep:`0492`. Requires privileged entity.
            If set to ``None`` (default), the setting will be untouched. Only the "global"
            notification setting is supported (ie, per client type is not possible).
        """
        existing = await self.__get_bookmark() if preserve else None

        new = Item()
        new["id"] = self.jid
        new["conference"]["nick"] = self.user_nick

        if existing is None:
            change = True
            new["conference"]["autojoin"] = auto_join
        else:
            change = False
            new["conference"]["autojoin"] = existing["conference"]["autojoin"]

        existing_extensions = existing is not None and existing[
            "conference"
        ].get_plugin("extensions", check=True)

        # preserving extensions we don't know about is a MUST
        if existing_extensions:
            assert existing is not None
            for el in existing["conference"]["extensions"].xml:
                if el.tag.startswith(f"{{{NOTIFY_NS}}}"):
                    if notify is not None:
                        continue
                if el.tag.startswith(f"{{{PINNING_NS}}}"):
                    if pin is not None:
                        continue
                new["conference"]["extensions"].append(el)

        if pin is not None:
            if existing_extensions:
                assert existing is not None
                existing_pin = (
                    existing["conference"]["extensions"].get_plugin(
                        "pinned", check=True
                    )
                    is not None
                )
                if existing_pin != pin:
                    change = True
            new["conference"]["extensions"]["pinned"] = pin

        if notify is not None:
            new["conference"]["extensions"].enable("notify")
            if existing_extensions:
                assert existing is not None
                existing_notify = existing["conference"]["extensions"].get_plugin(
                    "notify", check=True
                )
                if existing_notify is None:
                    change = True
                else:
                    if existing_notify.get_config() != notify:
                        change = True
                    for el in existing_notify:
                        new["conference"]["extensions"]["notify"].append(el)
            new["conference"]["extensions"]["notify"].configure(notify)

        if change:
            iq = Iq(stype="set", sfrom=self.user_jid, sto=self.user_jid)
            iq["pubsub"]["publish"]["node"] = self.xmpp["xep_0402"].stanza.NS
            iq["pubsub"]["publish"].append(new)

            iq["pubsub"]["publish_options"] = _BOOKMARKS_OPTIONS

            try:
                await self.xmpp["xep_0356"].send_privileged_iq(iq)
            except PermissionError:
                warnings.warn(
                    "IQ privileges (XEP0356) are not set, we cannot add bookmarks for the user"
                )
                # fallback by forcing invitation
                bookmark_add_fail = True
            except IqError as e:
                warnings.warn(
                    f"Something went wrong while trying to set the bookmarks: {e}"
                )
                # fallback by forcing invitation
                bookmark_add_fail = True
            else:
                bookmark_add_fail = False
        else:
            self.log.debug("Bookmark does not need updating.")
            return

        if bookmark_add_fail:
            self.session.send_gateway_invite(
                self,
                reason="This group could not be added automatically for you, most"
                "likely because this gateway is not configured as a privileged entity. "
                "Contact your administrator.",
            )
        elif existing is None and self.session.user.preferences.get(
            "always_invite_when_adding_bookmarks", True
        ):
            self.session.send_gateway_invite(
                self,
                reason="The gateway is configured to always send invitations for groups.",
            )

    async def on_avatar(self, data: bytes | None, mime: str | None) -> int | str | None:
        """
        Called when the user tries to set the avatar of the room from an XMPP
        client.

        If the set avatar operation is completed, should return a legacy image
        unique identifier. In this case the MUC avatar will be immediately
        updated on the XMPP side.

        If data is not None and this method returns None, then we assume that
        self.set_avatar() will be called elsewhere, eg triggered by a legacy
        room update event.

        :param data: image data or None if the user meant to remove the avatar
        :param mime: the mime type of the image. Since this is provided by
            the XMPP client, there is no guarantee that this is valid or
            correct.
        :return: A unique avatar identifier, which will trigger
            :py:meth:`slidge.group.room.LegacyMUC.set_avatar`. Alternatively, None, if
            :py:meth:`.LegacyMUC.set_avatar` is meant to be awaited somewhere else.
        """
        raise NotImplementedError

    admin_set_avatar = deprecated("LegacyMUC.on_avatar", on_avatar)

    async def on_set_affiliation(
        self,
        contact: "LegacyContact",
        affiliation: MucAffiliation,
        reason: str | None,
        nickname: str | None,
    ):
        """
        Triggered when the user requests changing the affiliation of a contact
        for this group.

        Examples: promotion them to moderator, ban (affiliation=outcast).

        :param contact: The contact whose affiliation change is requested
        :param affiliation: The new affiliation
        :param reason: A reason for this affiliation change
        :param nickname:
        """
        raise NotImplementedError

    async def on_kick(self, contact: "LegacyContact", reason: str | None):
        """
        Triggered when the user requests changing the role of a contact
        to "none" for this group. Action commonly known as "kick".

        :param contact: Contact to be kicked
        :param reason: A reason for this kick
        """
        raise NotImplementedError

    async def on_set_config(
        self,
        name: str | None,
        description: str | None,
    ):
        """
        Triggered when the user requests changing the room configuration.
        Only title and description can be changed at the moment.

        The legacy module is responsible for updating :attr:`.title` and/or
        :attr:`.description` of this instance.

        If :attr:`.HAS_DESCRIPTION` is set to False, description will always
        be ``None``.

        :param name: The new name of the room.
        :param description: The new description of the room.
        """
        raise NotImplementedError

    async def on_destroy_request(self, reason: str | None):
        """
        Triggered when the user requests room destruction.

        :param reason: Optionally, a reason for the destruction
        """
        raise NotImplementedError

    async def parse_mentions(self, text: str) -> list[Mention]:
        with self.xmpp.store.session() as orm:
            await self.__fill_participants()
            orm.add(self.stored)
            participants = {p.nickname: p for p in self.stored.participants}

            if len(participants) == 0:
                return []

            result = []
            for match in re.finditer(
                "|".join(
                    sorted(
                        [re.escape(nick) for nick in participants.keys()],
                        key=lambda nick: len(nick),
                        reverse=True,
                    )
                ),
                text,
            ):
                span = match.span()
                nick = match.group()
                if span[0] != 0 and text[span[0] - 1] not in _WHITESPACE_OR_PUNCTUATION:
                    continue
                if span[1] == len(text) or text[span[1]] in _WHITESPACE_OR_PUNCTUATION:
                    participant = self.participant_from_store(
                        stored=participants[nick],
                    )
                    if contact := participant.contact:
                        result.append(
                            Mention(contact=contact, start=span[0], end=span[1])
                        )
        return result

    async def on_set_subject(self, subject: str) -> None:
        """
        Triggered when the user requests changing the room subject.

        The legacy module is responsible for updating :attr:`.subject` of this
        instance.

        :param subject: The new subject for this room.
        """
        raise NotImplementedError

    async def on_set_thread_subject(
        self, thread: LegacyThreadType, subject: str
    ) -> None:
        """
        Triggered when the user requests changing the subject of a specific thread.

        :param thread: Legacy identifier of the thread
        :param subject: The new subject for this thread.
        """
        raise NotImplementedError

    @property
    def participants_filled(self) -> bool:
        return self.stored.participants_filled

    def get_archived_messages(
        self, msg_id: LegacyMessageType | str
    ) -> Iterator[HistoryMessage]:
        """
        Query the slidge archive for messages sent in this group

        :param msg_id: Message ID of the message in question. Can be either a legacy ID
            or an XMPP ID.
        :return: Iterator over messages. A single legacy ID can map to several messages,
            because of multi-attachment messages.
        """
        with self.xmpp.store.session() as orm:
            for stored in self.xmpp.store.mam.get_messages(
                orm, self.stored.id, ids=[str(msg_id)]
            ):
                yield HistoryMessage(stored.stanza)


def set_origin_id(msg: Message, origin_id: str) -> None:
    sub = ET.Element("{urn:xmpp:sid:0}origin-id")
    sub.attrib["id"] = origin_id
    msg.xml.append(sub)


def int_or_none(x):
    try:
        return int(x)
    except ValueError:
        return None


def equals_zero(x):
    if x is None:
        return False
    else:
        return x == 0


def str_to_datetime_or_none(date: str | None):
    if date is None:
        return
    try:
        return str_to_datetime(date)
    except ValueError:
        return None


def bookmarks_form():
    form = Form()
    form["type"] = "submit"
    form.add_field(
        "FORM_TYPE",
        value="http://jabber.org/protocol/pubsub#publish-options",
        ftype="hidden",
    )
    form.add_field("pubsub#persist_items", value="1")
    form.add_field("pubsub#max_items", value="max")
    form.add_field("pubsub#send_last_published_item", value="never")
    form.add_field("pubsub#access_model", value="whitelist")
    return form


_BOOKMARKS_OPTIONS = bookmarks_form()
_WHITESPACE_OR_PUNCTUATION = string.whitespace + "!\"'(),.:;?@_"

log = logging.getLogger(__name__)



"""Add unique constraint on avatar URL

Revision ID: abea9b63fa34
Revises: 805fdc0bf984
Create Date: 2026-02-09 20:20:53.576218

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "abea9b63fa34"
down_revision: str | None = "805fdc0bf984"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    op.execute("""
        DELETE FROM avatar
        WHERE id NOT IN (
            SELECT MIN(id)
            FROM avatar
            GROUP BY url
        );
    """)

    with op.batch_alter_table("avatar", schema=None) as batch_op:
        batch_op.create_unique_constraint(batch_op.f("uq_avatar_url"), ["url"])


def downgrade() -> None:
    with op.batch_alter_table("avatar", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("uq_avatar_url"), type_="unique")



"""
The main slidge package.

Contains importable classes for a minimal function :term:`Legacy Module`.
"""

import warnings
from importlib.metadata import PackageNotFoundError, version
from typing import Any

from . import slixfix  # noqa: F401
from .command import FormField, SearchResult  # noqa: F401
from .contact import LegacyContact, LegacyRoster  # noqa: F401
from .core import config as global_config  # noqa: F401
from .core.gateway import BaseGateway  # noqa: F401
from .core.session import BaseSession  # noqa: F401
from .db import GatewayUser  # noqa: F401
from .group import LegacyBookmarks, LegacyMUC, LegacyParticipant  # noqa: F401
from .main import main as main_func
from .util.types import MucType  # noqa: F401
from .util.util import addLoggingLevel


def entrypoint(module_name: str) -> None:
    """
    Entrypoint to be used in ``__main__.py`` of
    :term:`legacy modules <Legacy Module>`.

    :param module_name: An importable :term:`Legacy Module`.
    """
    main_func(module_name)


def formatwarning(
    message: Any, category: Any, filename: Any, lineno: Any, line: str = ""
) -> str:
    return f"{filename}:{lineno}:{category.__name__}:{message}\n"


warnings.formatwarning = formatwarning  # type:ignore

try:
    __version__ = version("slidge")
except PackageNotFoundError:
    # package is not installed
    __version__ = "dev"


__all__ = [
    "__version__",
    "BaseGateway",
    "BaseSession",
    # For backwards compatibility, these names are still importable from the
    # top-level slidge module, but this is deprecated.
    # "GatewayUser",
    # "LegacyBookmarks",
    # "LegacyMUC",
    # "LegacyContact",
    # "LegacyParticipant",
    # "LegacyRoster",
    # "MucType",
    # "FormField",
    # "SearchResult",
    "entrypoint",
    "global_config",
]

addLoggingLevel()



"""Initial schema

Revision ID: cef02a8b1451
Revises:
Create Date: 2025-08-28 12:48:16.890606

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

import slidge

# revision identifiers, used by Alembic.
revision: str = "cef02a8b1451"
down_revision: str | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "avatar",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("hash", sa.String(), nullable=False),
        sa.Column("height", sa.Integer(), nullable=False),
        sa.Column("width", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=True),
        sa.Column("etag", sa.String(), nullable=True),
        sa.Column("last_modified", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_avatar")),
        sa.UniqueConstraint("hash", name=op.f("uq_avatar_hash")),
        sa.UniqueConstraint("legacy_id", name=op.f("uq_avatar_legacy_id")),
    )
    op.create_table(
        "bob",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("file_name", sa.String(), nullable=False),
        sa.Column("sha_1", sa.String(), nullable=False),
        sa.Column("sha_256", sa.String(), nullable=False),
        sa.Column("sha_512", sa.String(), nullable=False),
        sa.Column("content_type", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_bob")),
        sa.UniqueConstraint("sha_1", name=op.f("uq_bob_sha_1")),
        sa.UniqueConstraint("sha_256", name=op.f("uq_bob_sha_256")),
        sa.UniqueConstraint("sha_512", name=op.f("uq_bob_sha_512")),
    )
    op.create_table(
        "user_account",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column(
            "registration_date",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.Column(
            "legacy_module_data", slidge.db.meta.JSONEncodedDict(), nullable=False
        ),
        sa.Column("preferences", slidge.db.meta.JSONEncodedDict(), nullable=False),
        sa.Column("avatar_hash", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_user_account")),
        sa.UniqueConstraint("jid", name=op.f("uq_user_account_jid")),
    )
    op.create_table(
        "attachment",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_file_id", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=False),
        sa.Column("sims", sa.String(), nullable=True),
        sa.Column("sfs", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
            name=op.f("fk_attachment_user_account_id_user_account"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_attachment")),
        sa.UniqueConstraint(
            "user_account_id",
            "legacy_file_id",
            name="uq_attachment_user_account_id_legacy_file_id",
        ),
    )
    with op.batch_alter_table("attachment", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_attachment_legacy_file_id"), ["legacy_file_id"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_attachment_url"), ["url"], unique=False)

    op.create_table(
        "contact",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column("avatar_id", sa.Integer(), nullable=True),
        sa.Column("nick", sa.String(), nullable=True),
        sa.Column("cached_presence", sa.Boolean(), nullable=False),
        sa.Column("last_seen", sa.DateTime(), nullable=True),
        sa.Column("ptype", sa.String(), nullable=True),
        sa.Column("pstatus", sa.String(), nullable=True),
        sa.Column("pshow", sa.String(), nullable=True),
        sa.Column("caps_ver", sa.String(), nullable=True),
        sa.Column("is_friend", sa.Boolean(), nullable=False),
        sa.Column("added_to_roster", sa.Boolean(), nullable=False),
        sa.Column("extra_attributes", slidge.db.meta.JSONEncodedDict(), nullable=True),
        sa.Column("updated", sa.Boolean(), nullable=False),
        sa.Column("vcard", sa.String(), nullable=True),
        sa.Column("vcard_fetched", sa.Boolean(), nullable=False),
        sa.Column(
            "client_type",
            sa.Enum(
                "bot",
                "console",
                "game",
                "handheld",
                "pc",
                "phone",
                "sms",
                "tablet",
                "web",
                native_enum=False,
            ),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["avatar_id"], ["avatar.id"], name=op.f("fk_contact_avatar_id_avatar")
        ),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
            name=op.f("fk_contact_user_account_id_user_account"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_contact")),
        sa.UniqueConstraint(
            "user_account_id", "jid", name="uq_contact_user_account_id_jid"
        ),
        sa.UniqueConstraint(
            "user_account_id", "legacy_id", name="uq_contact_user_account_id_legacy_id"
        ),
    )
    op.create_table(
        "room",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column("avatar_id", sa.Integer(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("subject", sa.String(), nullable=True),
        sa.Column("subject_date", sa.DateTime(), nullable=True),
        sa.Column("subject_setter", sa.String(), nullable=True),
        sa.Column("n_participants", sa.Integer(), nullable=True),
        sa.Column(
            "muc_type",
            sa.Enum("GROUP", "CHANNEL", "CHANNEL_NON_ANONYMOUS", name="muctype"),
            nullable=False,
        ),
        sa.Column("user_nick", sa.String(), nullable=True),
        sa.Column("user_resources", sa.String(), nullable=True),
        sa.Column("participants_filled", sa.Boolean(), nullable=False),
        sa.Column("history_filled", sa.Boolean(), nullable=False),
        sa.Column("extra_attributes", slidge.db.meta.JSONEncodedDict(), nullable=True),
        sa.Column("updated", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["avatar_id"], ["avatar.id"], name=op.f("fk_room_avatar_id_avatar")
        ),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
            name=op.f("fk_room_user_account_id_user_account"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_room")),
        sa.UniqueConstraint(
            "user_account_id", "jid", name="uq_room_user_account_id_jid"
        ),
        sa.UniqueConstraint(
            "user_account_id", "legacy_id", name="uq_room_user_account_id_legacy_id"
        ),
    )
    op.create_table(
        "contact_sent",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("contact_id", sa.Integer(), nullable=False),
        sa.Column("msg_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["contact_id"],
            ["contact.id"],
            name=op.f("fk_contact_sent_contact_id_contact"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_contact_sent")),
        sa.UniqueConstraint(
            "contact_id", "msg_id", name="uq_contact_sent_contact_id_msg_id"
        ),
    )
    op.create_table(
        "direct_msg",
        sa.Column("foreign_key", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("xmpp_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["foreign_key"],
            ["contact.id"],
            name=op.f("fk_direct_msg_foreign_key_contact"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_direct_msg")),
    )
    with op.batch_alter_table("direct_msg", schema=None) as batch_op:
        batch_op.create_index(
            "ix_direct_msg_legacy_id", ["legacy_id", "foreign_key"], unique=False
        )

    op.create_table(
        "direct_thread",
        sa.Column("foreign_key", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("xmpp_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["foreign_key"],
            ["contact.id"],
            name=op.f("fk_direct_thread_foreign_key_contact"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_direct_thread")),
    )
    with op.batch_alter_table("direct_thread", schema=None) as batch_op:
        batch_op.create_index(
            "ix_direct_direct_thread_id", ["legacy_id", "foreign_key"], unique=False
        )

    op.create_table(
        "group_msg",
        sa.Column("foreign_key", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("xmpp_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["foreign_key"], ["room.id"], name=op.f("fk_group_msg_foreign_key_room")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_group_msg")),
    )
    with op.batch_alter_table("group_msg", schema=None) as batch_op:
        batch_op.create_index(
            "ix_group_msg_legacy_id", ["legacy_id", "foreign_key"], unique=False
        )

    op.create_table(
        "group_thread",
        sa.Column("foreign_key", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("xmpp_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["foreign_key"], ["room.id"], name=op.f("fk_group_thread_foreign_key_room")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_group_thread")),
    )
    with op.batch_alter_table("group_thread", schema=None) as batch_op:
        batch_op.create_index(
            "ix_direct_group_thread_id", ["legacy_id", "foreign_key"], unique=False
        )

    op.create_table(
        "mam",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("room_id", sa.Integer(), nullable=False),
        sa.Column("stanza_id", sa.String(), nullable=False),
        sa.Column("timestamp", sa.DateTime(), nullable=False),
        sa.Column("author_jid", slidge.db.meta.JIDType(), nullable=False),
        sa.Column(
            "source",
            sa.Enum("LIVE", "BACKFILL", name="archivedmessagesource"),
            nullable=False,
        ),
        sa.Column("legacy_id", sa.String(), nullable=True),
        sa.Column("stanza", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["room_id"], ["room.id"], name=op.f("fk_mam_room_id_room")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_mam")),
        sa.UniqueConstraint("room_id", "stanza_id", name="uq_mam_room_id_stanza_id"),
    )
    op.create_table(
        "participant",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("room_id", sa.Integer(), nullable=False),
        sa.Column("contact_id", sa.Integer(), nullable=True),
        sa.Column("is_user", sa.Boolean(), nullable=False),
        sa.Column(
            "affiliation",
            sa.Enum("outcast", "member", "admin", "owner", "none", native_enum=False),
            nullable=False,
        ),
        sa.Column(
            "role",
            sa.Enum("moderator", "participant", "visitor", "none", native_enum=False),
            nullable=False,
        ),
        sa.Column("presence_sent", sa.Boolean(), nullable=False),
        sa.Column("resource", sa.String(), nullable=False),
        sa.Column("nickname", sa.String(), nullable=False),
        sa.Column("nickname_no_illegal", sa.String(), nullable=False),
        sa.Column("hats", sa.JSON(), nullable=False),
        sa.Column("extra_attributes", slidge.db.meta.JSONEncodedDict(), nullable=True),
        sa.ForeignKeyConstraint(
            ["contact_id"],
            ["contact.id"],
            name=op.f("fk_participant_contact_id_contact"),
        ),
        sa.ForeignKeyConstraint(
            ["room_id"], ["room.id"], name=op.f("fk_participant_room_id_room")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_participant")),
        sa.UniqueConstraint(
            "room_id", "contact_id", name="uq_participant_room_id_contact_id"
        ),
        sa.UniqueConstraint(
            "room_id", "resource", name="uq_participant_room_id_resource"
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("participant")
    op.drop_table("mam")
    with op.batch_alter_table("group_thread", schema=None) as batch_op:
        batch_op.drop_index("ix_direct_group_thread_id")

    op.drop_table("group_thread")
    with op.batch_alter_table("group_msg", schema=None) as batch_op:
        batch_op.drop_index("ix_group_msg_legacy_id")

    op.drop_table("group_msg")
    with op.batch_alter_table("direct_thread", schema=None) as batch_op:
        batch_op.drop_index("ix_direct_direct_thread_id")

    op.drop_table("direct_thread")
    with op.batch_alter_table("direct_msg", schema=None) as batch_op:
        batch_op.drop_index("ix_direct_msg_legacy_id")

    op.drop_table("direct_msg")
    op.drop_table("contact_sent")
    op.drop_table("room")
    op.drop_table("contact")
    with op.batch_alter_table("attachment", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_attachment_url"))
        batch_op.drop_index(batch_op.f("ix_attachment_legacy_file_id"))

    op.drop_table("attachment")
    op.drop_table("user_account")
    op.drop_table("bob")
    op.drop_table("avatar")
    # ### end Alembic commands ###



from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0153 import VCardTempUpdate, stanza
from slixmpp.stanza import Presence
from slixmpp.xmlstream import register_stanza_plugin


class XEP_0153(BasePlugin):
    name = "xep_0153"
    description = "XEP-0153: vCard-Based Avatars (slidge, just for MUCs)"
    dependencies = {"xep_0054"}
    stanza = stanza

    def plugin_init(self) -> None:
        register_stanza_plugin(Presence, VCardTempUpdate)



from unittest import mock

import pytest
from slixmpp import JID

from slidge.group import LegacyParticipant
from slidge.db.models import Participant


@pytest.fixture
def muc():
    muc = mock.MagicMock()
    muc.jid = JID("room@component")
    return muc


def test_unassigned_code_points(muc):
    part = LegacyParticipant(muc, Participant(nickname="fiesta! 🎉"))
    assert "🎉" not in part.jid.resource


def test_control_chars(muc):
    part = LegacyParticipant(
        muc, Participant(nickname="leet hackk\ber and I have control chars in my nick")
    )
    assert "\b" not in part.jid.resource


def test_control_chars_and_unassigned_code_points(muc):
    part = LegacyParticipant(
        muc,
        Participant(
            nickname="I'm a leet hackk\ber"
            + "🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉" * 10
            + ", I have control chars, emojis in my nick and a ridiculously long nickname"
        ),
    )
    assert "\b" not in part.jid.resource
    assert "🎉" not in part.jid.resource



from pathlib import Path

from slixmpp import JID as JIDType

# REQUIRED, so not default value


class _Categories:
    MANDATORY = (0, "Mandatory settings")
    BASE = (10, "Basic configuration")
    ATTACHMENTS = (20, "Attachments")
    LOG = (30, "Logging")
    ADVANCED = (40, "Advanced settings")


LEGACY_MODULE: str
LEGACY_MODULE__DOC = (
    "Importable python module containing (at least) a BaseGateway and a LegacySession subclass. "
    "NB: this is not needed if you use a gateway-specific entrypoint, e.g., `slidgram` or "
    "`python -m slidgram`."
)
LEGACY_MODULE__CATEGORY = _Categories.BASE

SERVER: str = "localhost"
SERVER__DOC = (
    "The XMPP server's host name. Defaults to localhost, which is the "
    "standard way of running slidge, on the same host as the XMPP server. "
    "The 'Jabber Component Protocol' (XEP-0114) does not mention encryption, "
    "so you *should* provide encryption another way, eg via port forwarding, if "
    "you change this."
)
SERVER__SHORT = "s"
SERVER__CATEGORY = _Categories.BASE

SECRET: str
SECRET__DOC = "The gateway component's secret (required to connect to the XMPP server)"
SECRET__CATEGORY = _Categories.MANDATORY

JID: JIDType
JID__DOC = "The gateway component's JID"
JID__SHORT = "j"
JID__CATEGORY = _Categories.MANDATORY

PORT: str = "5347"
PORT__DOC = "The XMPP server's port for incoming component connections"
PORT__SHORT = "p"
PORT__CATEGORY = _Categories.BASE

# Dynamic default (depends on other values)

HOME_DIR: Path
HOME_DIR__DOC = (
    "Directory where slidge will writes it persistent data and cache. "
    "Defaults to /var/lib/slidge/${SLIDGE_JID}. "
)
HOME_DIR__DYNAMIC_DEFAULT = True
HOME_DIR__CATEGORY = _Categories.BASE

DB_URL: str
DB_URL__DOC = (
    "Database URL, see <https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls>. "
    "Defaults to sqlite:///${HOME_DIR}/slidge.sqlite"
)
DB_URL__DYNAMIC_DEFAULT = True
DB_URL__CATEGORY = _Categories.ADVANCED

USER_JID_VALIDATOR: str
USER_JID_VALIDATOR__DOC = (
    "Regular expression to restrict users that can register to the gateway, by JID. "
    "Defaults to .*@${INFERRED_SERVER}. INFERRED_SERVER is derived for the gateway JID, "
    "by removing whatever is before the first encountered dot in it. Example: if "
    "slidge's JID=slidge.example.org, INFERRED_SERVER=example.org."
)
USER_JID_VALIDATOR__DYNAMIC_DEFAULT = True
USER_JID_VALIDATOR__CATEGORY = _Categories.BASE

# Optional, so default value + type hint if default is None

ADMINS: tuple[JIDType, ...] = ()
ADMINS__DOC = "JIDs of the gateway admins"
ADMINS__CATEGORY = _Categories.BASE

UPLOAD_SERVICE: str | None = None
UPLOAD_SERVICE__DOC = (
    "JID of an HTTP upload service the gateway can use. "
    "This is optional, as it should be automatically determined via service"
    "discovery."
)
UPLOAD_SERVICE__CATEGORY = _Categories.ATTACHMENTS

AVATAR_SIZE = 200
AVATAR_SIZE__DOC = (
    "Maximum image size (width and height), image ratio will be preserved"
)
AVATAR_SIZE__CATEGORY = _Categories.ADVANCED

USE_ATTACHMENT_ORIGINAL_URLS = False
USE_ATTACHMENT_ORIGINAL_URLS__DOC = (
    "For legacy plugins in which attachments are publicly downloadable URLs, "
    "let XMPP clients directly download them from this URL. Note that this will "
    "probably leak your client IP to the legacy network."
)
USE_ATTACHMENT_ORIGINAL_URLS__CATEGORY = _Categories.ATTACHMENTS

UPLOAD_REQUESTER: str | None = None
UPLOAD_REQUESTER__DOC = (
    "Set which JID should request the upload slots. Defaults to the user's JID if "
    "IQ/get privileges granted for the 'urn:xmpp:http:upload:0' namespace; the component "
    "JID otherwise."
)
UPLOAD_REQUESTER__CATEGORY = _Categories.ATTACHMENTS

UPLOAD_URL_PREFIX: str | None = None
UPLOAD_URL_PREFIX__DOC = (
    "This is an optional setting to make sure the URL of your upload service is never leaked "
    "to the legacy network in bodies of messages. This can happen under rare circumstances and/or bugs,"
    "when replying to an attachment. Set this to the common prefix of the public URL your attachments get, "
    "eg https://upload.example.org:5281/"
)
UPLOAD_URL_PREFIX__CATEGORY = _Categories.ATTACHMENTS

NO_UPLOAD_PATH: str | None = None
NO_UPLOAD_PATH__DOC = (
    "Instead of using the XMPP server's HTTP upload component, copy files to this dir. "
    "You need to set NO_UPLOAD_URL_PREFIX too if you use this option, and configure "
    "an web server to serve files in this dir."
)
NO_UPLOAD_PATH__CATEGORY = _Categories.ATTACHMENTS

NO_UPLOAD_URL_PREFIX: str | None = None
NO_UPLOAD_URL_PREFIX__DOC = (
    "Base URL that servers files in the dir set in the no-upload-path option, "
    "eg https://example.com:666/slidge-attachments/"
)
NO_UPLOAD_URL_PREFIX__CATEGORY = _Categories.ATTACHMENTS

NO_UPLOAD_METHOD: str = "copy"
NO_UPLOAD_METHOD__DOC = (
    "Whether to 'copy', 'move', 'hardlink' or 'symlink' the files in no-upload-path."
)
NO_UPLOAD_METHOD__CATEGORY = _Categories.ATTACHMENTS

NO_UPLOAD_FILE_READ_OTHERS = False
NO_UPLOAD_FILE_READ_OTHERS__DOC = (
    "After writing a file in NO_UPLOAD_PATH, change its permission so that 'others' can"
    " read it."
)
NO_UPLOAD_FILE_READ_OTHERS__CATEGORY = _Categories.ATTACHMENTS

IGNORE_DELAY_THRESHOLD = 300
IGNORE_DELAY_THRESHOLD__DOC = (
    "Threshold, in seconds, below which the <delay> information is stripped "
    "out of emitted stanzas."
)
IGNORE_DELAY_THRESHOLD__CATEGORY = _Categories.ADVANCED

PARTIAL_REGISTRATION_TIMEOUT = 3600
PARTIAL_REGISTRATION_TIMEOUT__DOC = (
    "Timeout before registration and login. Only useful for legacy networks where "
    "a single step registration process is not enough."
)
PARTIAL_REGISTRATION_TIMEOUT__CATEGORY = _Categories.ADVANCED

QR_TIMEOUT = 60
QR_TIMEOUT__DOC = "Timeout for QR code flashing confirmation."
QR_TIMEOUT__CATEGORY = _Categories.ADVANCED

FIX_FILENAME_SUFFIX_MIME_TYPE = False
FIX_FILENAME_SUFFIX_MIME_TYPE__DOC = (
    "Fix the Filename suffix based on the Mime Type of the file. Some clients (eg"
    " Conversations) may not inline files that have a wrong suffix for the MIME Type."
    " Therefore the MIME Type of the file is checked, if the suffix is not valid for"
    " that MIME Type, a valid one will be picked."
)
FIX_FILENAME_SUFFIX_MIME_TYPE__CATEGORY = _Categories.ATTACHMENTS

LOG_FILE: Path | None = None
LOG_FILE__DOC = "Log to a file instead of stdout/err"
LOG_FILE__CATEGORY = _Categories.LOG

LOG_FORMAT: str = "%(levelname)s:%(name)s:%(message)s"
LOG_FORMAT__DOC = (
    "Optionally, a format string for logging messages. Refer to "
    "https://docs.python.org/3/library/logging.html#logrecord-attributes "
    "for available options."
)
LOG_FORMAT__CATEGORY = _Categories.LOG

MAM_MAX_DAYS = 7
MAM_MAX_DAYS__DOC = "Maximum number of days for group archive retention."
MAM_MAX_DAYS__CATEGORY = _Categories.BASE

ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH = 200
ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH__DOC = (
    "Some legacy network provide ridiculously long filenames, strip above this limit, "
    "preserving suffix."
)
ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH__CATEGORY = _Categories.ATTACHMENTS

CONVERT_STICKERS = False
CONVERT_STICKERS__DOC = (
    "Convert lottie vector stickers (from the legacy side) to webp animations."
)
CONVERT_STICKERS__CATEGORY = _Categories.ATTACHMENTS

AVATAR_RESAMPLING_THREADS = 2
AVATAR_RESAMPLING_THREADS__DOC = (
    "Number of additional threads to use for avatar resampling. Even in a single-core "
    "context, this makes avatar resampling non-blocking."
)
AVATAR_RESAMPLING_THREADS__CATEGORY = _Categories.ADVANCED

DEV_MODE = False
DEV_MODE__DOC = (
    "Enables an interactive python shell via chat commands, for admins."
    "Not safe to use in prod, but great during dev."
)
DEV_MODE__CATEGORY = _Categories.ADVANCED


STRIP_LEADING_EMOJI_ADHOC = False
STRIP_LEADING_EMOJI_ADHOC__DOC = (
    "Strip the leading emoji in ad-hoc command names, if present, in case you "
    "are a emoji-hater."
)
STRIP_LEADING_EMOJI_ADHOC__CATEGORY = _Categories.ADVANCED

COMPONENT_NAME: str | None = None
COMPONENT_NAME__DOC = (
    "Overrides the default component name with a custom one. This is seen in service discovery and as the nickname "
    "of the component in chat windows."
)
COMPONENT_NAME__CATEGORY = _Categories.ADVANCED

WELCOME_MESSAGE: str | None = None
WELCOME_MESSAGE__DOC = (
    "Overrides the default welcome message received by newly registered users."
)
WELCOME_MESSAGE__CATEGORY = _Categories.ADVANCED



import abc
import asyncio
import logging
from collections.abc import Iterable
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    NamedTuple,
    Optional,
    Union,
    cast,
)

import aiohttp
import sqlalchemy as sa
from slixmpp import JID, Message
from slixmpp.exceptions import XMPPError
from slixmpp.types import PresenceShows, ResourceDict

from ..command import SearchResult
from ..contact import LegacyContact, LegacyRoster
from ..db.models import Contact, GatewayUser
from ..group.bookmarks import LegacyBookmarks
from ..group.room import LegacyMUC
from ..util import SubclassableOnce
from ..util.lock import NamedLockMixin
from ..util.types import (
    LegacyGroupIdType,
    LegacyMessageType,
    LegacyThreadType,
    LinkPreview,
    Mention,
    PseudoPresenceShow,
    RecipientType,
    Sticker,
)
from ..util.util import deprecated, noop_coro

if TYPE_CHECKING:
    from ..group.participant import LegacyParticipant
    from ..util.types import Sender
    from .gateway import BaseGateway


class CachedPresence(NamedTuple):
    status: str | None
    show: str | None
    kwargs: dict[str, Any]


class BaseSession(
    Generic[LegacyMessageType, RecipientType],
    NamedLockMixin,
    SubclassableOnce,
    abc.ABC,
):
    """
    The session of a registered :term:`User`.

    Represents a gateway user logged in to the legacy network and performing actions.

    Will be instantiated automatically on slidge startup for each registered user,
    or upon registration for new (validated) users.

    Must be subclassed for a functional :term:`Legacy Module`.
    """

    """
    Since we cannot set the XMPP ID of messages sent by XMPP clients, we need to keep a mapping
    between XMPP IDs and legacy message IDs if we want to further refer to a message that was sent
    by the user. This also applies to 'carboned' messages, ie, messages sent by the user from
    the official client of a legacy network.
    """

    xmpp: "BaseGateway"
    """
    The gateway instance singleton. Use it for low-level XMPP calls or custom methods that are not
    session-specific.
    """

    MESSAGE_IDS_ARE_THREAD_IDS = False
    """
    Set this to True if the legacy service uses message IDs as thread IDs,
    eg Mattermost, where you can only 'create a thread' by replying to the message,
    in which case the message ID is also a thread ID (and all messages are potential
    threads).
    """
    SPECIAL_MSG_ID_PREFIX: str | None = None
    """
    If you set this, XMPP message IDs starting with this won't be converted to legacy ID,
    but passed as is to :meth:`.on_react`, and usual checks for emoji restriction won't be
    applied.
    This can be used to implement voting in polls in a hacky way.
    """

    _roster_cls: type[LegacyRoster]
    _bookmarks_cls: type[LegacyBookmarks]

    def __init__(self, user: GatewayUser) -> None:
        super().__init__()
        self.user = user
        self.log = logging.getLogger(user.jid.bare)

        self.ignore_messages = set[str]()

        self.contacts = self._roster_cls(self)
        self.is_logging_in = False
        self._logged = False
        self.__reset_ready()

        self.bookmarks = self._bookmarks_cls(self)

        self.thread_creation_lock = asyncio.Lock()

        self.__cached_presence: CachedPresence | None = None

        self.__tasks = set[asyncio.Task]()

    @property
    def user_jid(self) -> JID:
        return self.user.jid

    @property
    def user_pk(self) -> int:
        return self.user.id

    @property
    def http(self) -> aiohttp.ClientSession:
        return self.xmpp.http

    def __remove_task(self, fut) -> None:
        self.log.debug("Removing fut %s", fut)
        self.__tasks.remove(fut)

    def create_task(self, coro, name: str | None = None) -> asyncio.Task:
        task = self.xmpp.loop.create_task(coro, name=name)
        self.__tasks.add(task)
        self.log.debug("Creating task %s", task)
        task.add_done_callback(lambda _: self.__remove_task(task))
        return task

    def cancel_all_tasks(self) -> None:
        for task in self.__tasks:
            task.cancel()

    @abc.abstractmethod
    async def login(self) -> str | None:
        """
        Logs in the gateway user to the legacy network.

        Triggered when the gateway start and on user registration.
        It is recommended that this function returns once the user is logged in,
        so if you need to await forever (for instance to listen to incoming events),
        it's a good idea to wrap your listener in an asyncio.Task.

        :return: Optionally, a text to use as the gateway status, e.g., "Connected as 'dude@legacy.network'"
        """
        raise NotImplementedError

    async def logout(self) -> None:
        """
        Logs out the gateway user from the legacy network.

        Called on gateway shutdown.
        """
        raise NotImplementedError

    async def on_text(
        self,
        chat: RecipientType,
        text: str,
        *,
        reply_to_msg_id: LegacyMessageType | None = None,
        reply_to_fallback_text: str | None = None,
        reply_to: Optional["Sender"] = None,
        thread: LegacyThreadType | None = None,
        link_previews: Iterable[LinkPreview] = (),
        mentions: list[Mention] | None = None,
    ) -> LegacyMessageType | None:
        """
        Triggered when the user sends a text message from XMPP to a bridged entity, e.g.
        to ``translated_user_name@slidge.example.com``, or ``translated_group_name@slidge.example.com``

        Override this and implement sending a message to the legacy network in this method.

        :param text: Content of the message
        :param chat: Recipient of the message. :class:`.LegacyContact` instance for 1:1 chat,
            :class:`.MUC` instance for groups.
        :param reply_to_msg_id: A legacy message ID if the message references (quotes)
            another message (:xep:`0461`)
        :param reply_to_fallback_text: Content of the quoted text. Not necessarily set
            by XMPP clients
        :param reply_to: Author of the quoted message. :class:`LegacyContact` instance for
            1:1 chat, :class:`LegacyParticipant` instance for groups.
            If `None`, should be interpreted as a self-reply if reply_to_msg_id is not None.
        :param link_previews: A list of sender-generated link previews.
            At the time of writing, only `Cheogram <https://wiki.soprani.ca/CheogramApp/LinkPreviews>`_
            supports it.
        :param mentions: (only for groups) A list of Contacts mentioned by their
            nicknames.
        :param thread:

        :return: An ID of some sort that can be used later to ack and mark the message
            as read by the user
        """
        raise NotImplementedError

    send_text = deprecated("BaseSession.send_text", on_text)

    async def on_file(
        self,
        chat: RecipientType,
        url: str,
        *,
        http_response: aiohttp.ClientResponse,
        reply_to_msg_id: LegacyMessageType | None = None,
        reply_to_fallback_text: str | None = None,
        reply_to: Union["LegacyContact", "LegacyParticipant"] | None = None,
        thread: LegacyThreadType | None = None,
    ) -> LegacyMessageType | None:
        """
        Triggered when the user sends a file using HTTP Upload (:xep:`0363`)

        :param url: URL of the file
        :param chat: See :meth:`.BaseSession.on_text`
        :param http_response: The HTTP GET response object on the URL
        :param reply_to_msg_id: See :meth:`.BaseSession.on_text`
        :param reply_to_fallback_text: See :meth:`.BaseSession.on_text`
        :param reply_to: See :meth:`.BaseSession.on_text`
        :param thread:

        :return: An ID of some sort that can be used later to ack and mark the message
            as read by the user
        """
        raise NotImplementedError

    send_file = deprecated("BaseSession.send_file", on_file)

    async def on_sticker(
        self,
        chat: RecipientType,
        sticker: Sticker,
        *,
        reply_to_msg_id: LegacyMessageType | None = None,
        reply_to_fallback_text: str | None = None,
        reply_to: Union["LegacyContact", "LegacyParticipant"] | None = None,
        thread: LegacyThreadType | None = None,
    ) -> LegacyMessageType | None:
        """
        Triggered when the user sends a file using HTTP Upload (:xep:`0363`)

        :param chat: See :meth:`.BaseSession.on_text`
        :param sticker: The sticker sent by the user.
        :param reply_to_msg_id: See :meth:`.BaseSession.on_text`
        :param reply_to_fallback_text: See :meth:`.BaseSession.on_text`
        :param reply_to: See :meth:`.BaseSession.on_text`
        :param thread:

        :return: An ID of some sort that can be used later to ack and mark the message
            as read by the user
        """
        raise NotImplementedError

    async def on_active(
        self, chat: RecipientType, thread: LegacyThreadType | None = None
    ):
        """
        Triggered when the user sends an 'active' chat state (:xep:`0085`)

        :param chat: See :meth:`.BaseSession.on_text`
        :param thread:
        """
        raise NotImplementedError

    active = deprecated("BaseSession.active", on_active)

    async def on_inactive(
        self, chat: RecipientType, thread: LegacyThreadType | None = None
    ):
        """
        Triggered when the user sends an 'inactive' chat state (:xep:`0085`)

        :param chat: See :meth:`.BaseSession.on_text`
        :param thread:
        """
        raise NotImplementedError

    inactive = deprecated("BaseSession.inactive", on_inactive)

    async def on_composing(
        self, chat: RecipientType, thread: LegacyThreadType | None = None
    ):
        """
        Triggered when the user starts typing in a legacy chat (:xep:`0085`)

        :param chat: See :meth:`.BaseSession.on_text`
        :param thread:
        """
        raise NotImplementedError

    composing = deprecated("BaseSession.composing", on_composing)

    async def on_paused(
        self, chat: RecipientType, thread: LegacyThreadType | None = None
    ):
        """
        Triggered when the user pauses typing in a legacy chat (:xep:`0085`)

        :param chat: See :meth:`.BaseSession.on_text`
        :param thread:
        """
        raise NotImplementedError

    paused = deprecated("BaseSession.paused", on_paused)

    async def on_gone(
        self, chat: RecipientType, thread: LegacyThreadType | None = None
    ):
        """
        Triggered when the user is "gone" in a legacy chat (:xep:`0085`)

        :param chat: See :meth:`.BaseSession.on_text`
        :param thread:
        """
        raise NotImplementedError

    async def on_displayed(
        self,
        chat: RecipientType,
        legacy_msg_id: LegacyMessageType,
        thread: LegacyThreadType | None = None,
    ):
        """
        Triggered when the user reads a message in a legacy chat. (:xep:`0333`)

        This is only possible if a valid ``legacy_msg_id`` was passed when
        transmitting a message from a legacy chat to the user, eg in
        :meth:`slidge.contact.LegacyContact.send_text`
        or
        :meth:`slidge.group.LegacyParticipant.send_text`.

        :param chat: See :meth:`.BaseSession.on_text`
        :param legacy_msg_id: Identifier of the message/
        :param thread:
        """
        raise NotImplementedError

    displayed = deprecated("BaseSession.displayed", on_displayed)

    async def on_correct(
        self,
        chat: RecipientType,
        text: str,
        legacy_msg_id: LegacyMessageType,
        *,
        thread: LegacyThreadType | None = None,
        link_previews: Iterable[LinkPreview] = (),
        mentions: list[Mention] | None = None,
    ) -> LegacyMessageType | None:
        """
        Triggered when the user corrects a message using :xep:`0308`

        This is only possible if a valid ``legacy_msg_id`` was returned by
        :meth:`.on_text`.

        :param chat: See :meth:`.BaseSession.on_text`
        :param text: The new text
        :param legacy_msg_id: Identifier of the edited message
        :param thread:
        :param link_previews: A list of sender-generated link previews.
            At the time of writing, only `Cheogram <https://wiki.soprani.ca/CheogramApp/LinkPreviews>`_
            supports it.
        :param mentions: (only for groups) A list of Contacts mentioned by their
            nicknames.
        """
        raise NotImplementedError

    correct = deprecated("BaseSession.correct", on_correct)

    async def on_react(
        self,
        chat: RecipientType,
        legacy_msg_id: LegacyMessageType,
        emojis: list[str],
        thread: LegacyThreadType | None = None,
    ):
        """
        Triggered when the user sends message reactions (:xep:`0444`).

        :param chat: See :meth:`.BaseSession.on_text`
        :param thread:
        :param legacy_msg_id: ID of the message the user reacts to
        :param emojis: Unicode characters representing reactions to the message ``legacy_msg_id``.
            An empty string means "no reaction", ie, remove all reactions if any were present before
        """
        raise NotImplementedError

    react = deprecated("BaseSession.react", on_react)

    async def on_retract(
        self,
        chat: RecipientType,
        legacy_msg_id: LegacyMessageType,
        thread: LegacyThreadType | None = None,
    ):
        """
        Triggered when the user retracts (:xep:`0424`) a message.

        :param chat: See :meth:`.BaseSession.on_text`
        :param thread:
        :param legacy_msg_id: Legacy ID of the retracted message
        """
        raise NotImplementedError

    retract = deprecated("BaseSession.retract", on_retract)

    async def on_presence(
        self,
        resource: str,
        show: PseudoPresenceShow,
        status: str,
        resources: dict[str, ResourceDict],
        merged_resource: ResourceDict | None,
    ):
        """
        Called when the gateway component receives a presence, ie, when
        one of the user's clients goes online of offline, or changes its
        status.

        :param resource: The XMPP client identifier, arbitrary string.
        :param show: The presence ``<show>``, if available. If the resource is
            just 'available' without any ``<show>`` element, this is an empty
            str.
        :param status: A status message, like a deeply profound quote, eg,
            "Roses are red, violets are blue, [INSERT JOKE]".
        :param resources: A summary of all the resources for this user.
        :param merged_resource: A global presence for the user account,
            following rules described in :meth:`merge_resources`
        """
        raise NotImplementedError

    presence = deprecated("BaseSession.presence", on_presence)

    async def on_search(self, form_values: dict[str, str]) -> SearchResult | None:
        """
        Triggered when the user uses Jabber Search (:xep:`0055`) on the component

        Form values is a dict in which keys are defined in :attr:`.BaseGateway.SEARCH_FIELDS`

        :param form_values: search query, defined for a specific plugin by overriding
            in :attr:`.BaseGateway.SEARCH_FIELDS`
        :return:
        """
        raise NotImplementedError

    search = deprecated("BaseSession.search", on_search)

    async def on_avatar(
        self,
        bytes_: bytes | None,
        hash_: str | None,
        type_: str | None,
        width: int | None,
        height: int | None,
    ) -> None:
        """
        Triggered when the user uses modifies their avatar via :xep:`0084`.

        :param bytes_: The data of the avatar. According to the spec, this
            should always be a PNG, but some implementations do not respect
            that. If `None` it means the user has unpublished their avatar.
        :param hash_: The SHA1 hash of the avatar data. This is an identifier of
            the avatar.
        :param type_: The MIME type of the avatar.
        :param width: The width of the avatar image.
        :param height: The height of the avatar image.
        """
        raise NotImplementedError

    async def on_moderate(
        self, muc: LegacyMUC, legacy_msg_id: LegacyMessageType, reason: str | None
    ):
        """
        Triggered when the user attempts to retract a message that was sent in
        a MUC using :xep:`0425`.

        If retraction is not possible, this should raise the appropriate
        XMPPError with a human-readable message.

        NB: the legacy module is responsible for calling
        :method:`LegacyParticipant.moderate` when this is successful, because
        slidge will acknowledge the moderation IQ, but will not send the
        moderation message from the MUC automatically.

        :param muc: The MUC in which the message was sent
        :param legacy_msg_id: The legacy ID of the message to be retracted
        :param reason: Optionally, a reason for the moderation, given by the
            user-moderator.
        """
        raise NotImplementedError

    async def on_create_group(
        self, name: str, contacts: list[LegacyContact]
    ) -> LegacyGroupIdType:
        """
        Triggered when the user request the creation of a group via the
        dedicated :term:`Command`.

        :param name: Name of the group
        :param contacts: list of contacts that should be members of the group
        """
        raise NotImplementedError

    async def on_invitation(
        self, contact: LegacyContact, muc: LegacyMUC, reason: str | None
    ) -> None:
        """
        Triggered when the user invites a :term:`Contact` to a legacy MUC via
        :xep:`0249`.

        The default implementation calls :meth:`LegacyMUC.on_set_affiliation`
        with the 'member' affiliation. Override if you want to customize this
        behaviour.

        :param contact: The invitee
        :param muc: The group
        :param reason: Optionally, a reason
        """
        await muc.on_set_affiliation(contact, "member", reason, None)

    async def on_leave_group(self, muc_legacy_id: LegacyGroupIdType):
        """
        Triggered when the user leaves a group via the dedicated slidge command
        or the :xep:`0077` ``<remove />`` mechanism.

        This should be interpreted as definitely leaving the group.

        :param muc_legacy_id: The legacy ID of the group to leave
        """
        raise NotImplementedError

    async def on_preferences(
        self, previous: dict[str, Any], new: dict[str, Any]
    ) -> None:
        """
        This is called when the user updates their preferences.

        Override this if you need set custom preferences field and need to trigger
        something when a preference has changed.
        """
        raise NotImplementedError

    def __reset_ready(self) -> None:
        self.ready = self.xmpp.loop.create_future()

    @property
    def logged(self):
        return self._logged

    @logged.setter
    def logged(self, v: bool) -> None:
        self.is_logging_in = False
        self._logged = v
        if self.ready.done():
            if v:
                return
            self.__reset_ready()
            self.shutdown(logout=False)
            with self.xmpp.store.session() as orm:
                self.xmpp.store.mam.reset_source(orm)
                self.xmpp.store.rooms.reset_updated(orm)
                self.xmpp.store.contacts.reset_updated(orm)
                orm.commit()
        else:
            if v:
                self.ready.set_result(True)

    def __repr__(self) -> str:
        return f"<Session of {self.user_jid}>"

    def shutdown(self, logout: bool = True) -> asyncio.Task:
        for m in self.bookmarks:
            m.shutdown()
        with self.xmpp.store.session() as orm:
            for jid in orm.execute(
                sa.select(Contact.jid).filter_by(user=self.user, is_friend=True)
            ).scalars():
                pres = self.xmpp.make_presence(
                    pfrom=jid,
                    pto=self.user_jid,
                    ptype="unavailable",
                    pstatus="Gateway has shut down.",
                )
                pres.send()
        if logout:
            return self.xmpp.loop.create_task(self.__logout())
        else:
            return self.xmpp.loop.create_task(noop_coro())

    async def __logout(self) -> None:
        try:
            await self.logout()
        except NotImplementedError:
            pass
        except KeyboardInterrupt:
            pass

    @staticmethod
    def legacy_to_xmpp_msg_id(legacy_msg_id: LegacyMessageType) -> str:
        """
        Convert a legacy msg ID to a valid XMPP msg ID.
        Needed for read marks, retractions and message corrections.

        The default implementation just converts the legacy ID to a :class:`str`,
        but this should be overridden in case some characters needs to be escaped,
        or to add some additional,
        :term:`legacy network <Legacy Network`>-specific logic.

        :param legacy_msg_id:
        :return: A string that is usable as an XMPP stanza ID
        """
        return str(legacy_msg_id)

    legacy_msg_id_to_xmpp_msg_id = staticmethod(
        deprecated("BaseSession.legacy_msg_id_to_xmpp_msg_id", legacy_to_xmpp_msg_id)
    )

    @staticmethod
    def xmpp_to_legacy_msg_id(i: str) -> LegacyMessageType:
        """
        Convert a legacy XMPP ID to a valid XMPP msg ID.
        Needed for read marks and message corrections.

        The default implementation just converts the legacy ID to a :class:`str`,
        but this should be overridden in case some characters needs to be escaped,
        or to add some additional,
        :term:`legacy network <Legacy Network`>-specific logic.

        The default implementation is an identity function.

        :param i: The XMPP stanza ID
        :return: An ID that can be used to identify a message on the legacy network
        """
        return cast(LegacyMessageType, i)

    xmpp_msg_id_to_legacy_msg_id = staticmethod(
        deprecated("BaseSession.xmpp_msg_id_to_legacy_msg_id", xmpp_to_legacy_msg_id)
    )

    def raise_if_not_logged(self):
        if not self.logged:
            raise XMPPError(
                "internal-server-error",
                text="You are not logged to the legacy network",
            )

    @classmethod
    def _from_user_or_none(cls, user):
        if user is None:
            log.debug("user not found")
            raise XMPPError(text="User not found", condition="subscription-required")

        session = _sessions.get(user.jid.bare)
        if session is None:
            _sessions[user.jid.bare] = session = cls(user)
        return session

    @classmethod
    def from_user(cls, user):
        return cls._from_user_or_none(user)

    @classmethod
    def from_stanza(cls, s) -> "BaseSession":
        # """
        # Get a user's :class:`.LegacySession` using the "from" field of a stanza
        #
        # Meant to be called from :class:`BaseGateway` only.
        #
        # :param s:
        # :return:
        # """
        return cls.from_jid(s.get_from())

    @classmethod
    def from_jid(cls, jid: JID) -> "BaseSession":
        # """
        # Get a user's :class:`.LegacySession` using its jid
        #
        # Meant to be called from :class:`BaseGateway` only.
        #
        # :param jid:
        # :return:
        # """
        session = _sessions.get(jid.bare)
        if session is not None:
            return session
        with cls.xmpp.store.session() as orm:
            user = orm.query(GatewayUser).filter_by(jid=jid.bare).one_or_none()
        return cls._from_user_or_none(user)

    @classmethod
    async def kill_by_jid(cls, jid: JID) -> None:
        # """
        # Terminate a user session.
        #
        # Meant to be called from :class:`BaseGateway` only.
        #
        # :param jid:
        # :return:
        # """
        log.debug("Killing session of %s", jid)
        for user_jid, session in _sessions.items():
            if user_jid == jid.bare:
                break
        else:
            log.debug("Did not find a session for %s", jid)
            return
        for c in session.contacts:
            c.unsubscribe()
        for m in session.bookmarks:
            m.shutdown()

        try:
            session = _sessions.pop(jid.bare)
        except KeyError:
            log.warning("User not found during unregistration")
            return

        session.cancel_all_tasks()

        await cls.xmpp.unregister(session)
        with cls.xmpp.store.session() as orm:
            orm.delete(session.user)
            orm.commit()

    def __ack(self, msg: Message) -> None:
        if not self.xmpp.PROPER_RECEIPTS:
            self.xmpp.delivery_receipt.ack(msg)

    def send_gateway_status(
        self,
        status: str | None = None,
        show=PresenceShows | None,
        **kwargs,
    ) -> None:
        """
        Send a presence from the gateway to the user.

        Can be used to indicate the user session status, ie "SMS code required", "connected", …

        :param status: A status message
        :param show: Presence stanza 'show' element. I suggest using "dnd" to show
            that the gateway is not fully functional
        """
        self.__cached_presence = CachedPresence(status, show, kwargs)
        self.xmpp.send_presence(
            pto=self.user_jid.bare, pstatus=status, pshow=show, **kwargs
        )

    def send_cached_presence(self, to: JID) -> None:
        if not self.__cached_presence:
            self.xmpp.send_presence(pto=to, ptype="unavailable")
            return
        self.xmpp.send_presence(
            pto=to,
            pstatus=self.__cached_presence.status,
            pshow=self.__cached_presence.show,
            **self.__cached_presence.kwargs,
        )

    def send_gateway_message(self, text: str, **msg_kwargs) -> None:
        """
        Send a message from the gateway component to the user.

        Can be used to indicate the user session status, ie "SMS code required", "connected", …

        :param text: A text
        """
        self.xmpp.send_text(text, mto=self.user_jid, **msg_kwargs)

    def send_gateway_invite(
        self,
        muc: LegacyMUC,
        reason: str | None = None,
        password: str | None = None,
    ) -> None:
        """
        Send an invitation to join a MUC, emanating from the gateway component.

        :param muc:
        :param reason:
        :param password:
        """
        self.xmpp.invite_to(muc, reason=reason, password=password, mto=self.user_jid)

    async def input(self, text: str, **msg_kwargs):
        """
        Request user input via direct messages from the gateway component.

        Wraps call to :meth:`.BaseSession.input`

        :param text: The prompt to send to the user
        :param msg_kwargs: Extra attributes
        :return:
        """
        return await self.xmpp.input(self.user_jid, text, **msg_kwargs)

    async def send_qr(self, text: str) -> None:
        """
        Sends a QR code generated from 'text' via HTTP Upload and send the URL to
        ``self.user``

        :param text: Text to encode as a QR code
        """
        await self.xmpp.send_qr(text, mto=self.user_jid)

    async def get_contact_or_group_or_participant(self, jid: JID, create: bool = True):
        if (contact := self.contacts.by_jid_only_if_exists(jid)) is not None:
            return contact
        if (muc := self.bookmarks.by_jid_only_if_exists(JID(jid.bare))) is not None:
            return await self.__get_muc_or_participant(muc, jid)
        else:
            muc = None

        if not create:
            return None

        try:
            return await self.contacts.by_jid(jid)
        except XMPPError:
            if muc is None:
                try:
                    muc = await self.bookmarks.by_jid(jid)
                except XMPPError:
                    return
            return await self.__get_muc_or_participant(muc, jid)

    @staticmethod
    async def __get_muc_or_participant(muc: LegacyMUC, jid: JID):
        if nick := jid.resource:
            return await muc.get_participant(nick, create=False, fill_first=True)
        return muc

    async def wait_for_ready(self, timeout: int | float | None = 10):
        # """
        # Wait until session, contacts and bookmarks are ready
        #
        # (slidge internal use)
        #
        # :param timeout:
        # :return:
        # """
        try:
            await asyncio.wait_for(asyncio.shield(self.ready), timeout)
            await asyncio.wait_for(asyncio.shield(self.contacts.ready), timeout)
            await asyncio.wait_for(asyncio.shield(self.bookmarks.ready), timeout)
        except TimeoutError:
            raise XMPPError(
                "recipient-unavailable",
                "Legacy session is not fully initialized, retry later",
            )

    def legacy_module_data_update(self, data: dict) -> None:
        user = self.user
        user.legacy_module_data.update(data)
        self.xmpp.store.users.update(user)

    def legacy_module_data_set(self, data: dict) -> None:
        user = self.user
        user.legacy_module_data = data
        self.xmpp.store.users.update(user)

    def legacy_module_data_clear(self) -> None:
        user = self.user
        user.legacy_module_data.clear()
        self.xmpp.store.users.update(user)


# keys = user.jid.bare
_sessions: dict[str, BaseSession] = {}
log = logging.getLogger(__name__)



import pytest
from slixmpp import JID, ComponentXMPP
from slixmpp.plugins.xep_0050.adhoc import XEP_0050

import slidge.command.adhoc
from slidge.command import Command, Form
from slidge.command.adhoc import AdhocProvider
from slidge.command.base import FormField
from slidge.util.test import SlixTestPlus


class MockSession:
    def __init__(self, jid):
        # self.jid = jid
        self.logged = True


@pytest.fixture(autouse=True)
def mock(monkeypatch, MockRE):
    monkeypatch.setattr(
        slidge.command.base, "is_admin", lambda j: j.user.startswith("admin")
    )
    monkeypatch.setattr(Command, "_get_session", lambda s, j: MockSession(j))
    monkeypatch.setattr(XEP_0050, "new_session", lambda _: "session-id")
    monkeypatch.setattr(
        ComponentXMPP,
        "jid_validator",
        MockRE,
        raising=False,
    )


class Command1(Command):
    NAME = "Command number one"
    NODE = "command1"

    async def run(self, session, ifrom):
        return Form(
            title="A title",
            instructions="Some instructions",
            fields=[
                FormField(
                    "jid",
                    type="jid-single",
                    label="Enter a JID",
                    value="user@host",
                    required=True,
                ),
                FormField(
                    "option",
                    type="list-single",
                    options=[
                        {"label": "Option 1", "value": "option1"},
                        {"label": "Option 2", "value": "option2"},
                    ],
                ),
            ],
            handler=self.finish,
            handler_kwargs={"arg1": "argument 1"},
        )

    @staticmethod
    async def finish(form_values, _session, ifrom, arg1):
        if form_values["jid"] == "bad@bad":
            raise RuntimeError("IT'S BAD, WE'RE FUCKED")
        assert isinstance(form_values["jid"], JID)
        return f"all good mate, {arg1}"


class Command2(Command):
    NAME = "Command number two"
    NODE = "command2"

    async def run(self, session, ifrom):
        return Form(
            title="A title",
            instructions="Some instructions",
            fields=[
                FormField(
                    "str",
                    type="list-multi",
                    options=[
                        {"label": "Option 1", "value": "option1"},
                        {"label": "Option 2", "value": "option2"},
                    ],
                ),
            ],
            handler=self.finish,
            handler_kwargs={"arg1": "argument 1"},
        )

    @staticmethod
    async def finish(form_values, _session, ifrom, arg1):
        assert isinstance(form_values["str"], list)
        assert all(isinstance(f, str) for f in form_values["str"])
        return f"all good mate, {arg1}"


class TestCommandsResults(SlixTestPlus):
    def setUp(self):
        super().setUp()
        self.stream_start(
            mode="component",
            plugins=["xep_0050"],
            jid="slidge.whatever.ass",
            server="whatever.ass",
        )
        self.adhoc = AdhocProvider(self.xmpp)
        self.adhoc.register(Command1(self.xmpp))
        self.adhoc.register(Command2(self.xmpp))

    def test_form_ok(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>A title</title>
                  <instructions>Some instructions</instructions>
                  <field var="jid"
                         type="jid-single"
                         label="Enter a JID">
                    <value>user@host</value>
                    <required />
                  </field>
                  <field var="option"
                         type="list-single">
                    <value />
                    <option label="Option 1">
                      <value>option1</value>
                    </option>
                    <option label="Option 2">
                      <value>option2</value>
                    </option>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       sessionid="session-id">
                <x xmlns='jabber:x:data'
                   type='submit'>
                  <field var='jid'>
                    <value>value@value</value>
                  </field>
                  <field var='option'>
                    <value>option1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="completed">
                <note type="info">all good mate, argument 1</note>
              </command>
            </iq>
            """
        )

    def test_form_bad_option(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>A title</title>
                  <instructions>Some instructions</instructions>
                  <field var="jid"
                         type="jid-single"
                         label="Enter a JID">
                    <value>user@host</value>
                    <required />
                  </field>
                  <field var="option"
                         type="list-single">
                    <value />
                    <option label="Option 1">
                      <value>option1</value>
                    </option>
                    <option label="Option 2">
                      <value>option2</value>
                    </option>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       sessionid="session-id">
                <x xmlns='jabber:x:data'
                   type='submit'>
                  <field var='jid'>
                    <value>value@value</value>
                  </field>
                  <field var='option'>
                    <value>option3</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <error type="modify">
                <not-acceptable xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Not a valid option: 'option3'</text>
              </error>
            </iq>
            """,
            use_values=False,
        )

    def test_form_exc(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>A title</title>
                  <instructions>Some instructions</instructions>
                  <field var="jid"
                         type="jid-single"
                         label="Enter a JID">
                    <value>user@host</value>
                    <required />
                  </field>
                  <field var="option"
                         type="list-single">
                    <value />
                    <option label="Option 1">
                      <value>option1</value>
                    </option>
                    <option label="Option 2">
                      <value>option2</value>
                    </option>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       sessionid="session-id">
                <x xmlns='jabber:x:data'
                   type='submit'>
                  <field var='jid'>
                    <value>bad@bad</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <error type="wait">
                <internal-server-error xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">IT'S BAD, WE'RE FUCKED</text>
              </error>
            </iq>
            """,
            use_values=False,
        )

    def test_form_bad_jid(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>A title</title>
                  <instructions>Some instructions</instructions>
                  <field var="jid"
                         type="jid-single"
                         label="Enter a JID">
                    <value>user@host</value>
                    <required />
                  </field>
                  <field var="option"
                         type="list-single">
                    <value />
                    <option label="Option 1">
                      <value>option1</value>
                    </option>
                    <option label="Option 2">
                      <value>option2</value>
                    </option>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       sessionid="session-id">
                <x xmlns='jabber:x:data'
                   type='submit'>
                  <field var='jid'>
                    <value>bad@bad@bad</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <error type="modify">
                <not-acceptable xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Not a valid JID: 'bad@bad@bad'</text>
              </error>
            </iq>
            """,
            use_values=False,
        )

    def test_multi(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command2'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>A title</title>
                  <instructions>Some instructions</instructions>
                  <field var="str"
                         type="list-multi">
                    <value />
                    <option label="Option 1">
                      <value>option1</value>
                    </option>
                    <option label="Option 2">
                      <value>option2</value>
                    </option>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       sessionid="session-id">
                <x xmlns='jabber:x:data'
                   type='submit'>
                  <field var='str'>
                    <value>option1</value>
                    <value>option2</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="session-id"
                       status="completed">
                <note type="info">all good mate, argument 1</note>
              </command>
            </iq>
            """
        )



import uuid
import warnings
from collections.abc import Iterable
from datetime import UTC, datetime
from typing import TYPE_CHECKING, cast

from slixmpp import JID, Message
from slixmpp.types import MessageTypes

from slidge.util import strip_illegal_chars

from ...db.models import GatewayUser
from ...slixfix.link_preview.stanza import LinkPreview as LinkPreviewStanza
from ...util.types import (
    ChatState,
    LegacyMessageType,
    LinkPreview,
    MessageReference,
    ProcessingHint,
)
from .. import config
from .base import BaseSender

if TYPE_CHECKING:
    from ...group import LegacyMUC, LegacyParticipant


class MessageMaker(BaseSender):
    mtype: MessageTypes = NotImplemented
    _can_send_carbon: bool = NotImplemented
    STRIP_SHORT_DELAY = False
    USE_STANZA_ID = False

    muc: "LegacyMUC"

    def _recipient_pk(self) -> int:
        return (
            self.muc.stored.id if self.is_participant else self.stored.id  # type:ignore
        )

    def _make_message(
        self,
        state: ChatState | None = None,
        hints: Iterable[ProcessingHint] = (),
        legacy_msg_id: LegacyMessageType | None = None,
        when: datetime | None = None,
        reply_to: MessageReference | None = None,
        carbon: bool = False,
        link_previews: Iterable[LinkPreview] | None = None,
        **kwargs,
    ):
        body = kwargs.pop("mbody", None)
        mfrom = kwargs.pop("mfrom", self.jid)
        mto = kwargs.pop("mto", None)
        thread = kwargs.pop("thread", None)
        if carbon and self._can_send_carbon:
            # the msg needs to have jabber:client as xmlns, so
            # we don't want to associate with the XML stream
            msg_cls = Message
        else:
            msg_cls = self.xmpp.Message  # type:ignore
        msg = msg_cls(
            sfrom=mfrom,
            stype=kwargs.pop("mtype", None) or self.mtype,
            sto=mto,
            **kwargs,
        )
        if body:
            msg["body"] = strip_illegal_chars(body, "�")
            state = "active"
        if thread:
            with self.xmpp.store.session() as orm:
                thread_str = str(thread)
                msg["thread"] = (
                    self.xmpp.store.id_map.get_thread(
                        orm,
                        self._recipient_pk(),
                        thread_str,
                        self.is_participant,
                    )
                    or thread_str
                )
        if state:
            msg["chat_state"] = state
        for hint in hints:
            msg.enable(hint)
        self._set_msg_id(msg, legacy_msg_id)
        self._add_delay(msg, when)
        if link_previews:
            self._add_link_previews(msg, link_previews)
        if reply_to:
            self._add_reply_to(msg, reply_to)
        return msg

    def _set_msg_id(
        self, msg: Message, legacy_msg_id: LegacyMessageType | None = None
    ) -> None:
        if legacy_msg_id is not None:
            i = self.session.legacy_to_xmpp_msg_id(legacy_msg_id)
            msg.set_id(i)
            if self.USE_STANZA_ID:
                msg["stanza_id"]["id"] = i
                msg["stanza_id"]["by"] = self.muc.jid  # type: ignore
        elif self.USE_STANZA_ID:
            msg["stanza_id"]["id"] = str(uuid.uuid4())
            msg["stanza_id"]["by"] = self.muc.jid  # type: ignore

    def _legacy_to_xmpp(self, legacy_id: LegacyMessageType) -> list[str]:
        with self.xmpp.store.session() as orm:
            ids = self.xmpp.store.id_map.get_xmpp(
                orm,
                self._recipient_pk(),
                str(legacy_id),
                self.is_participant,
            )
            if ids:
                return ids
        return [self.session.legacy_to_xmpp_msg_id(legacy_id)]

    def _add_delay(self, msg: Message, when: datetime | None) -> None:
        if when:
            if when.tzinfo is None:
                when = when.astimezone(UTC)
            if self.STRIP_SHORT_DELAY:
                delay = (datetime.now().astimezone(UTC) - when).seconds
                if delay < config.IGNORE_DELAY_THRESHOLD:
                    return
            msg["delay"].set_stamp(when)
            msg["delay"].set_from(self.xmpp.boundjid.bare)

    def _add_reply_to(self, msg: Message, reply_to: MessageReference) -> None:
        xmpp_id = self._legacy_to_xmpp(reply_to.legacy_id)[0]
        msg["reply"]["id"] = xmpp_id

        muc = getattr(self, "muc", None)

        if entity := reply_to.author:
            if entity == "user" or isinstance(entity, GatewayUser):
                if isinstance(entity, GatewayUser):
                    warnings.warn(
                        "Using a GatewayUser as the author of a "
                        "MessageReference is deprecated. Use the string 'user' "
                        "instead.",
                        DeprecationWarning,
                    )
                if muc:
                    jid = JID(muc.jid)
                    jid.resource = fallback_nick = muc.user_nick
                    msg["reply"]["to"] = jid
                else:
                    msg["reply"]["to"] = self.session.user_jid
                    # TODO: here we should use preferably use the PEP nick of the user
                    # (but it doesn't matter much)
                    fallback_nick = self.session.user_jid.user
            else:
                if muc:
                    if hasattr(entity, "muc"):
                        # TODO: accept a Contact here and use muc.get_participant_by_legacy_id()
                        # a bit of work because right now this is a sync function
                        entity = cast("LegacyParticipant", entity)
                        fallback_nick = entity.nickname
                    else:
                        warnings.warn(
                            "The author of a message reference in a MUC must be a"
                            " Participant instance, not a Contact"
                        )
                        fallback_nick = entity.name
                else:
                    fallback_nick = entity.name
                msg["reply"]["to"] = entity.jid
        else:
            fallback_nick = None

        if fallback := reply_to.body:
            msg["reply"].add_quoted_fallback(fallback, fallback_nick)

    @staticmethod
    def _add_link_previews(msg: Message, link_previews: Iterable[LinkPreview]) -> None:
        for preview in link_previews:
            element = LinkPreviewStanza()
            for i, name in enumerate(preview._fields):
                val = preview[i]
                if not val:
                    continue
                element[name] = val
            msg.append(element)



import logging
from typing import TYPE_CHECKING

from slixmpp import JID

if TYPE_CHECKING:
    from .. import BaseGateway


class YesSet(set):
    """
    A pseudo-set which always test True for membership
    """

    def __contains__(self, item) -> bool:
        log.debug("Test in")
        return True


class RosterBackend:
    """
    A pseudo-roster for the gateway component.

    If a user is in the user store, this will behave as if the user is part of the
    roster with subscription "both", and "none" otherwise.

    This is rudimentary but the only sane way I could come up with so far.
    """

    def __init__(self, xmpp: "BaseGateway") -> None:
        self.xmpp = xmpp

    @staticmethod
    def entries(_owner_jid, _default=None):
        return YesSet()

    @staticmethod
    def save(_owner_jid, _jid, _item_state, _db_state) -> None:
        pass

    def load(self, _owner_jid, jid, _db_state):
        session = self.xmpp.get_session_from_jid(JID(jid))
        if session is None:
            return {
                "name": "",
                "groups": [],
                "from": False,
                "to": False,
                "pending_in": False,
                "pending_out": False,
                "whitelisted": False,
                "subscription": "both",
            }
        else:
            return {
                "name": "",
                "groups": [],
                "from": True,
                "to": True,
                "pending_in": False,
                "pending_out": False,
                "whitelisted": False,
                "subscription": "none",
            }


log = logging.getLogger(__name__)



from alembic import context

from slidge import global_config
from slidge.db.meta import Base, get_engine

config = context.config

target_metadata = Base.metadata


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    try:
        # in prod
        connectable = get_engine(global_config.DB_URL)
    except AttributeError:
        # during dev, to generate migrations
        connectable = get_engine("sqlite+pysqlite:///dev/slidge.sqlite")

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            render_as_batch=True,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from slixmpp import CoroutineCallback, Iq, StanzaPath
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0004 import Form
from slixmpp.xmlstream import StanzaBase

from ..util import DispatcherMixin, exceptions_to_xmpp_errors


class MucOwnerMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp) -> None:
        super().__init__(xmpp)
        xmpp.register_handler(
            CoroutineCallback(
                "MUCOwnerGet",
                StanzaPath("iq@type=get/mucowner_query"),
                self.on_muc_owner_query,
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "MUCOwnerSet",
                StanzaPath("iq@type=set/mucowner_query"),
                self.on_muc_owner_set,
            )
        )

    @exceptions_to_xmpp_errors
    async def on_muc_owner_query(self, iq: StanzaBase) -> None:
        assert isinstance(iq, Iq)
        muc = await self.get_muc_from_stanza(iq)

        reply = iq.reply()

        form = Form(title="Slidge room configuration")
        form["instructions"] = (
            "Complete this form to modify the configuration of your room."
        )
        form.add_field(
            var="FORM_TYPE",
            type="hidden",
            value="http://jabber.org/protocol/muc#roomconfig",
        )
        form.add_field(
            var="muc#roomconfig_roomname",
            label="Natural-Language Room Name",
            type="text-single",
            value=muc.name,
        )
        if muc.HAS_DESCRIPTION:
            form.add_field(
                var="muc#roomconfig_roomdesc",
                label="Short Description of Room",
                type="text-single",
                value=muc.description,
            )

        muc_owner = iq["mucowner_query"]
        muc_owner.append(form)
        reply.append(muc_owner)
        reply.send()

    @exceptions_to_xmpp_errors
    async def on_muc_owner_set(self, iq: StanzaBase) -> None:
        assert isinstance(iq, Iq)
        muc = await self.get_muc_from_stanza(iq)
        query = iq["mucowner_query"]

        if form := query.get_plugin("form", check=True):
            values = form.get_values()
            await muc.on_set_config(
                name=values.get("muc#roomconfig_roomname"),
                description=(
                    values.get("muc#roomconfig_roomdesc")
                    if muc.HAS_DESCRIPTION
                    else None
                ),
            )
            form["type"] = "result"
            clear = False
        elif destroy := query.get_plugin("destroy", check=True):
            reason = destroy["reason"] or None
            await muc.on_destroy_request(reason)
            user_participant = await muc.get_user_participant()
            user_participant.stored.affiliation = "none"
            user_participant.stored.role = "none"
            presence = user_participant._make_presence(ptype="unavailable", force=True)
            presence["muc"].enable("destroy")
            if reason is not None:
                presence["muc"]["destroy"]["reason"] = reason
            user_participant._send(presence)
            await muc.session.bookmarks.remove(muc, kick=False)
            clear = True
        else:
            raise XMPPError("bad-request")

        iq.reply(clear=clear).send()



"""
Everything related to groups.
"""

from ..util.types import MucType
from .bookmarks import LegacyBookmarks
from .participant import LegacyParticipant
from .room import LegacyMUC

__all__ = ("LegacyBookmarks", "LegacyParticipant", "LegacyMUC", "MucType")



from slixmpp import register_stanza_plugin, Message
from slixmpp.plugins.xep_0333.stanza import Displayed
from slixmpp.plugins.xep_0334 import Store, NoStore, NoPermanentStore
from slixmpp.plugins.xep_0424.stanza import Retract

from slidge.group.archive import archivable
from slixmpp.test import SlixTest

class TestArchivable(SlixTest):
    def setUp(self):
        register_stanza_plugin(Message, Displayed)
        register_stanza_plugin(Message, Retract)
        register_stanza_plugin(Message, Store)
        register_stanza_plugin(Message, NoStore)
        register_stanza_plugin(Message, NoPermanentStore)

    def test_marker(self):
        msg = Message()
        assert not archivable(msg)
        msg.enable("displayed")
        assert archivable(msg)

    def test_retract(self):
        msg = Message()
        assert not archivable(msg)
        msg.enable("retract")
        assert archivable(msg)

    def test_hint_store(self):
        msg = Message()
        assert not archivable(msg)
        msg.enable("store")
        assert archivable(msg)

    def test_hint_no_store(self):
        msg = Message()
        msg["body"] = "boobobo"
        assert archivable(msg)

        msg.enable("no-store")
        assert not archivable(msg)

    def test_hint_no_permanent_store(self):
        msg = Message()
        msg["body"] = "boobobo"
        assert archivable(msg)

        msg.enable("no-permanent-store")
        assert not archivable(msg)



"""Add 'occupant_id' column to Participant

Revision ID: b7a06a86416a
Revises: 86d980dbb426
Create Date: 2025-12-14 21:39:45.025086

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "b7a06a86416a"
down_revision: str | None = "86d980dbb426"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    op.execute("DELETE FROM participant")
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("participant", schema=None) as batch_op:
        batch_op.add_column(sa.Column("occupant_id", sa.String(), nullable=False))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("participant", schema=None) as batch_op:
        batch_op.drop_column("occupant_id")
    # ### end Alembic commands ###



"""
Typing stuff
"""

from collections.abc import AsyncIterator, Hashable
from dataclasses import dataclass, fields
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NamedTuple,
    TypedDict,
    TypeVar,
    Union,
)

from slixmpp import Message, Presence
from slixmpp.types import PresenceShows, PresenceTypes, ResourceDict  # noqa: F401

if TYPE_CHECKING:
    from ..contact import LegacyContact
    from ..core.session import BaseSession
    from ..group import LegacyMUC
    from ..group.participant import LegacyParticipant

    AnyBaseSession = BaseSession[Any, Any]
else:
    # Hack to work around circular import.
    AnyBaseSession = int


LegacyGroupIdType = TypeVar("LegacyGroupIdType", bound=Hashable)
"""
Type of the unique identifier for groups, usually a str or an int,
but anything hashable should work.
"""
LegacyMessageType = TypeVar("LegacyMessageType", bound=Hashable)
LegacyThreadType = TypeVar("LegacyThreadType", bound=Hashable)
LegacyUserIdType = TypeVar("LegacyUserIdType", bound=Hashable)

LegacyContactType = TypeVar("LegacyContactType", bound="LegacyContact[Any]")
LegacyMUCType = TypeVar("LegacyMUCType", bound="LegacyMUC[Any, Any, Any, Any]")
LegacyParticipantType = TypeVar("LegacyParticipantType", bound="LegacyParticipant")

SessionType = TypeVar("SessionType", bound="BaseSession[Any, Any]")
Recipient = Union["LegacyMUC[Any, Any, Any, Any]", "LegacyContact[Any]"]
RecipientType = TypeVar("RecipientType", bound=Recipient)
Sender = Union["LegacyContact[Any]", "LegacyParticipant"]
LegacyFileIdType = int | str

ChatState = Literal["active", "composing", "gone", "inactive", "paused"]
ProcessingHint = Literal["no-store", "markable", "store"]
Marker = Literal["acknowledged", "received", "displayed"]
FieldType = Literal[
    "boolean",
    "fixed",
    "text-single",
    "text-multi",
    "jid-single",
    "jid-multi",
    "list-single",
    "list-multi",
    "text-private",
]
MucAffiliation = Literal["owner", "admin", "member", "outcast", "none"]
MucRole = Literal["visitor", "participant", "moderator", "none"]
# https://xmpp.org/registrar/disco-categories.html#client
ClientType = Literal[
    "bot", "console", "game", "handheld", "pc", "phone", "sms", "tablet", "web"
]
AttachmentDisposition = Literal["attachment", "inline"]


@dataclass
class MessageReference(Generic[LegacyMessageType]):
    """
    A "message reply", ie a "quoted message" (:xep:`0461`)

    At the very minimum, the legacy message ID attribute must be set, but to
    ensure that the quote is displayed in all XMPP clients, the author must also
    be set (use the string "user" if the slidge user is the author of the referenced
    message).
    The body is used as a fallback for XMPP clients that do not support :xep:`0461`
    of that failed to find the referenced message.
    """

    legacy_id: LegacyMessageType
    author: Union[Literal["user"], "LegacyParticipant", "LegacyContact"] | None = None
    body: str | None = None


@dataclass
class LegacyAttachment:
    """
    A file attachment to a message

    At the minimum, one of the ``path``, ``steam``, ``data`` or ``url`` attribute
    has to be set

    To be used with :meth:`.LegacyContact.send_files` or
    :meth:`.LegacyParticipant.send_files`
    """

    path: Path | str | None = None
    name: str | None = None
    stream: IO[bytes] | None = None
    aio_stream: AsyncIterator[bytes] | None = None
    data: bytes | None = None
    content_type: str | None = None
    legacy_file_id: str | int | None = None
    url: str | None = None
    caption: str | None = None
    disposition: AttachmentDisposition | None = None
    """
    A caption for this specific image. For a global caption for a list of attachments,
    use the ``body`` parameter of :meth:`.AttachmentMixin.send_files`
    """

    def __post_init__(self) -> None:
        if all(
            x is None
            for x in (self.path, self.stream, self.data, self.url, self.aio_stream)
        ):
            raise TypeError("There is not data in this attachment", self)
        if isinstance(self.path, str):
            self.path = Path(self.path)

    def format_for_user(self) -> str:
        if self.name:
            name = self.name
        elif self.path:
            name = self.path.name  # type:ignore[union-attr]
        elif self.url:
            name = self.url
        else:
            name = ""

        if self.caption:
            if name:
                name = f"{name}: {self.caption}"
            else:
                name = self.caption

        return name

    def __str__(self):
        attrs = ", ".join(
            f"{f.name}={getattr(self, f.name)!r}"
            for f in fields(self)
            if getattr(self, f.name) is not None and f.name != "data"
        )
        if self.data is not None:
            data_str = f"data=<{len(self.data)} bytes>"
            to_join = (attrs, data_str) if attrs else (data_str,)
            attrs = ", ".join(to_join)
        return f"Attachment({attrs})"


class MucType(IntEnum):
    """
    The type of group, private, public, anonymous or not.
    """

    GROUP = 0
    """
    A private group, members-only and non-anonymous, eg a family group.
    """
    CHANNEL = 1
    """
    A public group, aka an anonymous channel.
    """
    CHANNEL_NON_ANONYMOUS = 2
    """
    A public group where participants' legacy IDs are visible to everybody.
    """


PseudoPresenceShow = PresenceShows | Literal[""]


MessageOrPresenceTypeVar = TypeVar("MessageOrPresenceTypeVar", bound=Message | Presence)


class LinkPreview(NamedTuple):
    about: str
    title: str | None
    description: str | None
    url: str | None
    image: str | None
    type: str | None
    site_name: str | None


class Mention(NamedTuple):
    contact: "LegacyContact[Any]"
    start: int
    end: int


class Hat(NamedTuple):
    uri: str
    title: str
    hue: float | None = None


class UserPreferences(TypedDict):
    sync_avatar: bool
    sync_presence: bool


class MamMetadata(NamedTuple):
    id: str
    sent_on: datetime


class HoleBound(NamedTuple):
    id: int | str
    timestamp: datetime


class CachedPresence(NamedTuple):
    last_seen: datetime | None = None
    ptype: PresenceTypes | None = None
    pstatus: str | None = None
    pshow: PresenceShows | None = None


class Sticker(NamedTuple):
    path: Path
    content_type: str | None
    hashes: dict[str, str]


class Avatar(NamedTuple):
    path: Path | None = None
    unique_id: str | int | None = None
    url: str | None = None
    data: bytes | None = None



from slixmpp.exceptions import XMPPError

from slidge import LegacyContact, LegacyRoster
from slidge.util.types import Avatar

from .session import Session


class Roster(LegacyRoster[int, "Contact"]):
    async def fill(self):
        for i in 111, 222:
            yield await self.by_legacy_id(i)

    async def jid_username_to_legacy_id(self, jid_username: str) -> int:
        try:
            return int(jid_username)
        except ValueError:
            raise XMPPError(
                "bad-request", "This is not a valid username for this fake network"
            )


class Contact(LegacyContact[int]):
    session: "Session"

    async def update_info(self):
        profile = await self.session.legacy_client.get_profile(self.legacy_id)
        self.name = profile.nickname
        self.set_vcard(full_name=profile.full_name)
        await self.set_avatar(
            Avatar(path=profile.avatar, unique_id=profile.avatar_unique_id)
        )
        if self.legacy_id != 666:
            self.is_friend = True
        self.online()



from __future__ import annotations

import json
from typing import Any, Union

import sqlalchemy as sa
from slixmpp import JID
from sqlalchemy import Dialect


class JIDType(sa.TypeDecorator[JID]):
    """
    Custom SQLAlchemy type for JIDs
    """

    impl = sa.types.TEXT
    cache_ok = True

    def process_bind_param(self, value: JID | None, dialect: sa.Dialect) -> str | None:
        if value is None:
            return value
        return str(value)

    def process_result_value(
        self, value: str | None, dialect: sa.Dialect
    ) -> JID | None:
        if value is None:
            return value
        return JID(value)


JSONSerializableTypes = Union[str, float, None, "JSONSerializable"]
JSONSerializable = dict[str, JSONSerializableTypes]


class JSONEncodedDict(sa.TypeDecorator[JSONSerializable]):
    """
    Custom SQLAlchemy type for dictionaries stored as JSON

    Note that mutations of the dictionary are not detected by SQLAlchemy,
    which is why use ``attributes.flag_modified()`` in ``UserStore.update()``
    """

    impl = sa.VARCHAR

    cache_ok = True

    def process_bind_param(
        self, value: JSONSerializable | None, dialect: Dialect
    ) -> str | None:
        if value is None:
            return None
        return json.dumps(value)

    def process_result_value(
        self, value: Any | None, dialect: Dialect
    ) -> JSONSerializable | None:
        if value is None:
            return None
        return json.loads(value)  # type:ignore


class Base(sa.orm.DeclarativeBase):
    type_annotation_map = {JSONSerializable: JSONEncodedDict, JID: JIDType}


Base.metadata.naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_`%(constraint_name)s`",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


def get_engine(path: str, echo: bool = False, pool_size: int = 5) -> sa.Engine:
    from sqlalchemy import log as sqlalchemy_log

    engine = sa.create_engine(path, pool_size=pool_size)
    if echo:
        sqlalchemy_log._add_default_handler = lambda x: None  # type:ignore
        engine.echo = True
    return engine



import pytest
from slixmpp import JID
from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from slidge import GatewayUser
from slidge.db.meta import Base
from slidge.db.store import UserStore


@pytest.fixture
def engine():
    engine = create_engine("sqlite+pysqlite:///:memory:", echo=True)
    Base.metadata.create_all(engine)
    return engine


@pytest.fixture
def store(engine):
    yield UserStore(sessionmaker(engine))


def test_user(engine, store: UserStore):
    with Session(engine) as orm:
        user1 = GatewayUser(jid=JID("test-user@test-host"), legacy_module_data={})
        orm.add(user1)
        orm.commit()
        assert user1.jid == JID("test-user@test-host")

    user1.preferences = {"section": {"do_xxx": True}}
    store.update(user1)
    del user1

    with Session(engine) as orm:
        user2 = orm.query(GatewayUser).filter_by(jid=JID("test-user@test-host")).one()
        assert user2.preferences == {"section": {"do_xxx": True}}

    user2.preferences["section"]["do_xxx"] = False

    store.update(user2)
    with Session(engine) as orm:
        user3 = orm.query(GatewayUser).filter_by(jid=JID("test-user@test-host")).one()
        assert not user3.preferences["section"]["do_xxx"]



import pytest

from slidge.core.mixins.recipient import ReactionRecipientMixin


@pytest.mark.asyncio
async def test_no_restriction():
    x = ReactionRecipientMixin()
    assert await x.restricted_emoji_extended_feature() is None


@pytest.mark.asyncio
async def test_single_reaction_any_emoji():
    class X(ReactionRecipientMixin):
        REACTIONS_SINGLE_EMOJI = True

    x = X()
    form = await x.restricted_emoji_extended_feature()
    values = form.get_values()
    assert values["max_reactions_per_user"] == "1"
    assert values.get("allowlist") is None


@pytest.mark.asyncio
async def test_single_emoji():
    class X(ReactionRecipientMixin):
        async def available_emojis(self, legacy_msg_id=None):
            return "♥"

    x = X()
    form = await x.restricted_emoji_extended_feature()
    values = form.get_values()
    assert values.get("max_reactions_per_user") is None
    assert values.get("allowlist") == ["♥"]


@pytest.mark.asyncio
async def test_two_emojis():
    class X(ReactionRecipientMixin):
        async def available_emojis(self, legacy_msg_id=None):
            return "♥", "😛"

    x = X()
    form = await x.restricted_emoji_extended_feature()
    values = form.get_values()
    assert values.get("max_reactions_per_user") is None
    assert values.get("allowlist") == ["♥", "😛"]


@pytest.mark.asyncio
async def test_two_emojis_single_reaction():
    class X(ReactionRecipientMixin):
        REACTIONS_SINGLE_EMOJI = True

        async def available_emojis(self, legacy_msg_id=None):
            return "♥", "😛"

    x = X()
    form = await x.restricted_emoji_extended_feature()
    values = form.get_values()
    assert values.get("max_reactions_per_user") == "1"
    assert values.get("allowlist") == ["♥", "😛"]
